# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import subprocess

import mock
import pytest

from paasta_tools import firewall
from paasta_tools import firewall_update
from paasta_tools import yaml_tools as yaml
from paasta_tools.utils import TimeoutError


def test_parse_args_daemon():
    args = firewall_update.parse_args(
        [
            "-d",
            "mysoadir",
            "-v",
            "--synapse-service-dir",
            "myservicedir",
            "daemon",
            "-u",
            "123",
        ]
    )
    assert args.mode == "daemon"
    assert args.synapse_service_dir == "myservicedir"
    assert args.soa_dir == "mysoadir"
    assert args.update_secs == 123
    assert args.verbose


def test_parse_args_default_daemon():
    args = firewall_update.parse_args(["daemon"])
    assert args.mode == "daemon"
    assert args.synapse_service_dir == firewall.DEFAULT_SYNAPSE_SERVICE_DIR
    assert args.soa_dir == firewall_update.DEFAULT_SOA_DIR
    assert args.update_secs == firewall_update.DEFAULT_UPDATE_SECS
    assert not args.verbose


def test_parse_args_cron():
    args = firewall_update.parse_args(["-d", "mysoadir", "-v", "cron"])
    assert args.mode == "cron"
    assert args.soa_dir == "mysoadir"
    assert args.verbose


def test_parse_args_default_cron():
    args = firewall_update.parse_args(["cron"])
    assert args.mode == "cron"
    assert args.soa_dir == firewall_update.DEFAULT_SOA_DIR
    assert not args.verbose


@mock.patch(
    "paasta_tools.utils.load_system_paasta_config",
    autospec=True,
)
@mock.patch.object(
    firewall_update,
    "load_system_paasta_config",
    autospec=True,
    return_value=mock.Mock(**{"get_cluster.return_value": "mycluster"}),
)
@mock.patch.object(
    firewall,
    "services_running_here",
    autospec=True,
    return_value=(
        ("myservice", "hassecurityoutbound", "02:42:a9:fe:00:0a", "1.1.1.1"),
    ),
)
def test_smartstack_dependencies_of_running_firewalled_services(_, __, ___, tmpdir):
    soa_dir = tmpdir.mkdir("yelpsoa")
    myservice_dir = soa_dir.mkdir("myservice")

    kubernetes_config = {
        "hassecurityoutbound": {
            "dependencies_reference": "my_ref",
            "security": {"outbound_firewall": "block"},
        },
        "nosecurity": {"dependencies_reference": "my_ref"},
    }
    myservice_dir.join("kubernetes-mycluster.yaml").write(
        yaml.safe_dump(kubernetes_config)
    )

    dependencies_config = {
        "my_ref": [
            {"well-known": "internet"},
            {"smartstack": "mydependency.depinstance"},
            {"smartstack": "another.one"},
        ]
    }
    myservice_dir.join("dependencies.yaml").write(yaml.safe_dump(dependencies_config))

    result = firewall_update.smartstack_dependencies_of_running_firewalled_services(
        soa_dir=str(soa_dir)
    )
    assert dict(result) == {
        "mydependency.depinstance": {
            ("myservice", "hassecurityoutbound"),
        },
        "another.one": {
            ("myservice", "hassecurityoutbound"),
        },
    }


@mock.patch.object(
    firewall_update,
    "smartstack_dependencies_of_running_firewalled_services",
    autospec=True,
)
@mock.patch.object(
    firewall_update, "process_inotify_event", side_effect=StopIteration, autospec=True
)
def test_run_daemon(process_inotify_mock, smartstack_deps_mock, mock_daemon_args):
    class kill_after_too_long:
        def __init__(self):
            self.count = 0

        def __call__(self, *args, **kwargs):
            self.count += 1
            assert self.count <= 5, "Took too long to detect file change"
            return {}

    smartstack_deps_mock.side_effect = kill_after_too_long()
    subprocess.Popen(
        [
            "bash",
            "-c",
            "sleep 0.2; echo > %s/mydep.depinstance.json"
            % mock_daemon_args.synapse_service_dir,
        ]
    )
    with pytest.raises(StopIteration):
        firewall_update.run_daemon(mock_daemon_args)
    assert smartstack_deps_mock.call_count > 0
    assert process_inotify_mock.call_args[0][0][3] == b"mydep.depinstance.json"
    assert process_inotify_mock.call_args[0][1] == {}


@mock.patch.object(firewall, "firewall_flock", autospec=True)
@mock.patch.object(firewall, "general_update", autospec=True)
def test_run_cron(mock_general_update, mock_firewall_flock, mock_cron_args):
    firewall_update.run_cron(mock_cron_args)
    assert mock_general_update.called is True
    assert mock_firewall_flock.return_value.__enter__.called is True


@mock.patch.object(
    firewall, "firewall_flock", autospec=True, side_effect=TimeoutError("Oh noes")
)
@mock.patch.object(firewall, "general_update", autospec=True)
def test_run_cron_flock_error(mock_general_update, mock_firewall_flock, mock_cron_args):
    with pytest.raises(TimeoutError):
        firewall_update.run_cron(mock_cron_args)


@mock.patch.object(firewall_update, "log", autospec=True)
@mock.patch.object(firewall_update.firewall, "ensure_service_chains", autospec=True)
@mock.patch.object(firewall_update.firewall, "active_service_groups", autospec=True)
@mock.patch.object(firewall, "firewall_flock", autospec=True)
def test_process_inotify_event(
    firewall_flock_mock,
    active_service_groups_mock,
    ensure_service_chains_mock,
    log_mock,
):
    active_service_groups_mock.return_value = {
        firewall.ServiceGroup("myservice", "myinstance"): {"00:00:00:00:00:00"},
        firewall.ServiceGroup("anotherservice", "instance"): {"11:11:11:11:11:11"},
        firewall.ServiceGroup("thirdservice", "instance"): {"22:22:22:22:22:22"},
    }

    services_by_dependencies = {
        "mydep.depinstance": {
            ("myservice", "myinstance"),
            ("anotherservice", "instance"),
        }
    }
    soa_dir = mock.Mock()
    synapse_service_dir = mock.Mock()
    firewall_update.process_inotify_event(
        (None, None, None, b"mydep.depinstance.json"),
        services_by_dependencies,
        soa_dir,
        synapse_service_dir,
    )
    assert log_mock.debug.call_count == 3
    log_mock.debug.assert_any_call("Updated ('myservice', 'myinstance')")
    log_mock.debug.assert_any_call("Updated ('anotherservice', 'instance')")
    assert ensure_service_chains_mock.mock_calls == [
        mock.call(
            {
                firewall.ServiceGroup("myservice", "myinstance"): {"00:00:00:00:00:00"},
                firewall.ServiceGroup("anotherservice", "instance"): {
                    "11:11:11:11:11:11"
                },
            },
            soa_dir,
            synapse_service_dir,
        )
    ]

    assert firewall_flock_mock.return_value.__enter__.called is True

    # Verify that tmp writes do not apply
    log_mock.reset_mock()
    ensure_service_chains_mock.reset_mock()
    firewall_update.process_inotify_event(
        (None, None, None, b"mydep.depinstance.tmp"),
        services_by_dependencies,
        soa_dir,
        synapse_service_dir,
    )
    assert log_mock.debug.call_count == 1
    assert ensure_service_chains_mock.call_count == 0


@mock.patch.object(firewall_update, "log", autospec=True)
@mock.patch.object(firewall_update.firewall, "ensure_service_chains", autospec=True)
@mock.patch.object(firewall_update.firewall, "active_service_groups", autospec=True)
@mock.patch.object(
    firewall, "firewall_flock", autospec=True, side_effect=TimeoutError("Oh noes")
)
def test_process_inotify_event_flock_error(
    firewall_flock_mock,
    active_service_groups_mock,
    ensure_service_chains_mock,
    log_mock,
):
    active_service_groups_mock.return_value = {
        firewall.ServiceGroup("myservice", "myinstance"): {"00:00:00:00:00:00"},
        firewall.ServiceGroup("anotherservice", "instance"): {"11:11:11:11:11:11"},
        firewall.ServiceGroup("thirdservice", "instance"): {"22:22:22:22:22:22"},
    }

    services_by_dependencies = {
        "mydep.depinstance": {
            ("myservice", "myinstance"),
            ("anotherservice", "instance"),
        }
    }
    soa_dir = mock.Mock()
    synapse_service_dir = mock.Mock()
    firewall_update.process_inotify_event(
        (None, None, None, b"mydep.depinstance.json"),
        services_by_dependencies,
        soa_dir,
        synapse_service_dir,
    )
    assert log_mock.debug.call_count == 1
    assert log_mock.error.call_count == 1


@pytest.fixture
def mock_daemon_args(tmpdir):
    return firewall_update.parse_args(
        [
            "-d",
            str(tmpdir.mkdir("yelpsoa")),
            "--synapse-service-dir",
            str(tmpdir.mkdir("synapse")),
            "daemon",
        ]
    )


@pytest.fixture
def mock_cron_args(tmpdir):
    return firewall_update.parse_args(
        [
            "-d",
            str(tmpdir.mkdir("yelpsoa")),
            "--synapse-service-dir",
            str(tmpdir.mkdir("synapse")),
            "cron",
        ]
    )



# coding: utf-8

# flake8: noqa

"""
    Paasta API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.2.0
    Generated by: https://openapi-generator.tech
"""


__version__ = "1.0.0"

# import ApiClient
from paasta_tools.paastaapi.api_client import ApiClient

# import Configuration
from paasta_tools.paastaapi.configuration import Configuration

# import exceptions
from paasta_tools.paastaapi.exceptions import OpenApiException
from paasta_tools.paastaapi.exceptions import ApiAttributeError
from paasta_tools.paastaapi.exceptions import ApiTypeError
from paasta_tools.paastaapi.exceptions import ApiValueError
from paasta_tools.paastaapi.exceptions import ApiKeyError
from paasta_tools.paastaapi.exceptions import ApiException



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from json.decoder import JSONDecodeError

import mock

from paasta_tools.secret_tools import decrypt_secret
from paasta_tools.secret_tools import decrypt_secret_environment_for_service
from paasta_tools.secret_tools import decrypt_secret_environment_variables
from paasta_tools.secret_tools import decrypt_secret_volumes
from paasta_tools.secret_tools import get_hmac_for_secret
from paasta_tools.secret_tools import get_secret_hashes
from paasta_tools.secret_tools import get_secret_name_from_ref
from paasta_tools.secret_tools import get_secret_provider
from paasta_tools.secret_tools import is_secret_ref
from paasta_tools.secret_tools import SHARED_SECRET_SERVICE
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import SecretVolume


def test_is_secret_ref():
    assert is_secret_ref("SECRET(aaa-bbb-222_111)")
    assert not is_secret_ref("SECRET(#!$)")
    # herein is a lesson on how tests are hard:
    assert not is_secret_ref("anything_else")
    assert not is_secret_ref("")
    # this is just incase a non string leaks in somewhere
    # if it is not a string it can't be a secret ref
    # so this checks that we are catching the TypeError
    assert not is_secret_ref(None)
    assert not is_secret_ref(3)  # type: ignore


def test_is_secret_ref_shared():
    assert is_secret_ref("SHARED_SECRET(foo)")


def test_get_secret_name_from_ref():
    assert get_secret_name_from_ref("SECRET(aaa-bbb-222_111)") == "aaa-bbb-222_111"


def test_get_shared_secret_name_from_ref():
    assert (
        get_secret_name_from_ref("SHARED_SECRET(aaa-bbb-222_111)") == "aaa-bbb-222_111"
    )


def test_get_hmac_for_secret():
    with mock.patch(
        "paasta_tools.secret_tools.open", autospec=False
    ) as mock_open, mock.patch(
        "json.load", autospec=True
    ) as mock_json_load, mock.patch(
        "paasta_tools.secret_tools.get_secret_name_from_ref", autospec=True
    ) as mock_get_secret_name_from_ref:
        mock_json_load.return_value = {
            "environments": {"dev": {"signature": "notArealHMAC"}}
        }
        mock_get_secret_name_from_ref.return_value = "secretsquirrel"

        ret = get_hmac_for_secret(
            "SECRET(secretsquirrel)", "service-name", "/nail/blah", "dev"
        )
        mock_get_secret_name_from_ref.assert_called_with("SECRET(secretsquirrel)")
        mock_open.assert_called_with(
            "/nail/blah/service-name/secrets/secretsquirrel.json", "r"
        )
        assert ret == "notArealHMAC"

        ret = get_hmac_for_secret(
            "SECRET(secretsquirrel)", "service-name", "/nail/blah", "dev-what"
        )
        assert ret is None

        mock_open.side_effect = IOError
        ret = get_hmac_for_secret(
            "SECRET(secretsquirrel)", "service-name", "/nail/blah", "dev"
        )
        assert ret is None

        ret = get_hmac_for_secret(
            "SECRET(secretsquirrel)", "service-name", "/nail/blah", "dev"
        )
        assert ret is None

        mock_open.side_effect = None
        mock_json_load.side_effect = JSONDecodeError("", "", 1)
        ret = get_hmac_for_secret(
            "SECRET(secretsquirrel)", "service-name", "/nail/blah", "dev"
        )
        assert ret is None


def test_get_hmac_for_shared_secret():
    with mock.patch(
        "paasta_tools.secret_tools.open", autospec=False
    ) as mock_open, mock.patch(
        "json.load", autospec=True
    ) as mock_json_load, mock.patch(
        "paasta_tools.secret_tools.get_secret_name_from_ref", autospec=True
    ) as mock_get_secret_name_from_ref:
        mock_json_load.return_value = {
            "environments": {"dev": {"signature": "notArealHMAC"}}
        }
        mock_get_secret_name_from_ref.return_value = "secretsquirrel"

        ret = get_hmac_for_secret(
            "SHARED_SECRET(secretsquirrel)", "service-name", "/nail/blah", "dev"
        )
        mock_get_secret_name_from_ref.assert_called_with(
            "SHARED_SECRET(secretsquirrel)"
        )
        mock_open.assert_called_with(
            f"/nail/blah/{SHARED_SECRET_SERVICE}/secrets/secretsquirrel.json", "r"
        )
        assert ret == "notArealHMAC"


def test_get_secret_provider():
    with mock.patch(
        "paasta_tools.secret_providers.SecretProvider", autospec=True
    ) as mock_secret_provider:
        ret = get_secret_provider(
            secret_provider_name="paasta_tools.secret_providers",
            soa_dir="/nail/blah",
            service_name="test-service",
            cluster_names=["norcal-devc"],
            secret_provider_kwargs={"some": "thing"},
        )
        mock_secret_provider.assert_called_with(
            soa_dir="/nail/blah",
            service_name="test-service",
            cluster_names=["norcal-devc"],
            some="thing",
        )
        assert ret == mock_secret_provider.return_value


def test_get_secret_hashes():
    with mock.patch(
        "paasta_tools.secret_tools.is_secret_ref", autospec=True, return_value=False
    ) as mock_is_secret_ref, mock.patch(
        "paasta_tools.secret_tools.get_hmac_for_secret", autospec=True
    ) as mock_get_hmac_for_secret:
        env = {"SOME_VAR": "SOME_VAL"}

        assert get_secret_hashes(env, "dev", "service", DEFAULT_SOA_DIR) == {}
        mock_is_secret_ref.assert_called_with("SOME_VAL")
        assert not mock_get_hmac_for_secret.called

        mock_is_secret_ref.return_value = True
        expected = {"SOME_VAL": mock_get_hmac_for_secret.return_value}
        assert get_secret_hashes(env, "dev", "service", DEFAULT_SOA_DIR) == expected
        mock_is_secret_ref.assert_called_with("SOME_VAL")
        mock_get_hmac_for_secret.assert_called_with(
            env_var_val="SOME_VAL",
            service="service",
            soa_dir=DEFAULT_SOA_DIR,
            secret_environment="dev",
        )


@mock.patch("paasta_tools.secret_tools.is_secret_ref", autospec=True)
@mock.patch("paasta_tools.secret_tools.is_shared_secret", autospec=True)
@mock.patch(
    "paasta_tools.secret_tools.decrypt_secret_environment_for_service",
    autospec=True,
)
def test_decrypt_secret_environment_variables(
    mock_decrypt_for_service, mock_is_shared_secret, mock_is_secret_ref
):
    mock_environment = {
        "MY": "aaa",
        "SECRET": "SECRET(123)",
        "SECRET_SHARED": "SHARED_SECRET(abc)",
    }
    mock_is_secret_ref.side_effect = lambda val: "SECRET" in val
    mock_is_shared_secret.side_effect = lambda val: "SHARED" in val
    mock_decrypt_for_service.side_effect = [{"SECRET": "123"}, {"SECRET_SHARED": "abc"}]

    ret = decrypt_secret_environment_variables(
        secret_provider_name="vault",
        environment=mock_environment,
        soa_dir="/nail/blah",
        service_name="universe",
        cluster_name="mesosstage",
        secret_provider_kwargs={"some": "config"},
    )
    assert ret == {"SECRET": "123", "SECRET_SHARED": "abc"}

    assert mock_decrypt_for_service.call_args_list == [
        mock.call(
            {"SECRET": "SECRET(123)"},
            "universe",
            "vault",
            "/nail/blah",
            "mesosstage",
            {"some": "config", "vault_num_uses": 2},
        ),
        mock.call(
            {"SECRET_SHARED": "SHARED_SECRET(abc)"},
            SHARED_SECRET_SERVICE,
            "vault",
            "/nail/blah",
            "mesosstage",
            {"some": "config", "vault_num_uses": 2},
        ),
    ]


@mock.patch("paasta_tools.secret_tools.get_secret_provider", autospec=True)
def test_decrypt_secret_environment_for_service(mock_get_secret_provider):
    mock_secret_env = {"SECRET": "SECRET(123)"}
    mock_secret_provider = mock.Mock()
    mock_get_secret_provider.return_value = mock_secret_provider
    ret = decrypt_secret_environment_for_service(
        secret_env_vars=mock_secret_env,
        service_name="universe",
        secret_provider_name="vault",
        soa_dir="/nail/blah",
        cluster_name="mesosstage",
        secret_provider_kwargs={"some": "config"},
    )
    mock_get_secret_provider.assert_called_with(
        secret_provider_name="vault",
        soa_dir="/nail/blah",
        service_name="universe",
        cluster_names=["mesosstage"],
        secret_provider_kwargs={"some": "config"},
    )
    mock_secret_provider.decrypt_environment.assert_called_with(
        {"SECRET": "SECRET(123)"}
    )
    assert ret == mock_secret_provider.decrypt_environment.return_value


@mock.patch("paasta_tools.secret_tools.get_secret_provider", autospec=True)
def test_decrypt_secret_volumes_multiple_files(mock_get_secret_provider):
    mock_secret_provider = mock.Mock()
    mock_get_secret_provider.return_value = mock_secret_provider

    mock_secret_volumes_config = [
        SecretVolume(
            container_path="/the/container/path/",
            items=[
                {"key": "the_secret_name1", "path": "the_secret_filename1"},
                {"key": "the_secret_name2", "path": "the_secret_filename2"},
            ],
        )
    ]
    mock_secret_provider.decrypt_secret_raw.side_effect = [
        "the_secret_contents1",
        "the_secret_contents2",
    ]
    ret = decrypt_secret_volumes(
        secret_provider_name="vault",
        secret_volumes_config=mock_secret_volumes_config,
        soa_dir="/nail/blah",
        service_name="universe",
        cluster_name="mesosstage",
        secret_provider_kwargs={"some": "config"},
    )
    assert ret == {
        "/the/container/path/the_secret_filename1": "the_secret_contents1",
        "/the/container/path/the_secret_filename2": "the_secret_contents2",
    }


@mock.patch("paasta_tools.secret_tools.get_secret_provider", autospec=True)
def test_decrypt_secret_volumes_single_file(mock_get_secret_provider):
    mock_secret_provider = mock.Mock()
    mock_get_secret_provider.return_value = mock_secret_provider

    mock_secret_volumes_config = [
        SecretVolume(
            container_path="/the/container/path/",
            secret_name="the_secret_name",
        )
    ]
    mock_secret_provider.decrypt_secret_raw.side_effect = ["the_secret_contents"]
    ret = decrypt_secret_volumes(
        secret_provider_name="vault",
        secret_volumes_config=mock_secret_volumes_config,
        soa_dir="/nail/blah",
        service_name="universe",
        cluster_name="mesosstage",
        secret_provider_kwargs={"some": "config"},
    )
    assert ret == {
        "/the/container/path/the_secret_name": "the_secret_contents",
    }


@mock.patch("paasta_tools.secret_tools.get_secret_provider", autospec=True)
def test_decrypt_secret_decode_true(mock_get_secret_provider):
    mock_secret_provider = mock.Mock()
    mock_get_secret_provider.return_value = mock_secret_provider

    ret = decrypt_secret(
        secret_provider_name="vault",
        soa_dir="/nail/blah",
        service_name="universe",
        cluster_name="mesosstage",
        secret_provider_kwargs={"some": "config"},
        secret_name="the_secret_name",
        decode=True,
    )
    assert ret

    mock_secret_provider.decrypt_secret.assert_called()
    mock_secret_provider.decrypt_secret.assert_called_with("the_secret_name")


@mock.patch("paasta_tools.secret_tools.get_secret_provider", autospec=True)
def test_decrypt_secret_decode_false(mock_get_secret_provider):
    mock_secret_provider = mock.Mock()
    mock_get_secret_provider.return_value = mock_secret_provider

    ret = decrypt_secret(
        secret_provider_name="vault",
        soa_dir="/nail/blah",
        service_name="universe",
        cluster_name="mesosstage",
        secret_provider_kwargs={"some": "config"},
        secret_name="the_secret_name",
        decode=False,
    )
    assert ret

    mock_secret_provider.decrypt_secret_raw.assert_called()
    mock_secret_provider.decrypt_secret_raw.assert_called_with("the_secret_name")



# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import os
import socket
from typing import AbstractSet
from typing import Any
from typing import Collection
from typing import DefaultDict
from typing import Dict
from typing import FrozenSet
from typing import Iterable
from typing import List
from typing import Mapping
from typing import MutableMapping
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple

import requests
from kubernetes.client import V1Pod
from mypy_extensions import TypedDict

from paasta_tools import yaml_tools as yaml
from paasta_tools.utils import get_user_agent


class EnvoyBackend(TypedDict, total=False):
    address: str
    port_value: int
    hostname: str
    eds_health_status: str
    weight: int
    has_associated_task: bool


def are_services_up_in_pod(
    envoy_host: str,
    envoy_admin_port: int,
    envoy_admin_endpoint_format: str,
    registrations: Collection[str],
    pod_ip: str,
    pod_port: int,
) -> bool:
    """Returns whether a service in a k8s pod is reachable via envoy
    :param envoy_host: The host that this check should contact for replication information.
    :param envoy_admin_port: The port that Envoy's admin interface is listening on
    :param registrations: The service_name.instance_name of the services
    :param pod_ip: IP of the pod itself
    :param pod_port: The port to reach the service in the pod
    """

    for registration in registrations:
        backends_per_registration = get_backends(
            registration,
            envoy_host=envoy_host,
            envoy_admin_port=envoy_admin_port,
            envoy_admin_endpoint_format=envoy_admin_endpoint_format,
        )

        healthy_backends = [
            backend
            for backend in backends_per_registration.get(registration, [])
            if backend[0]["address"] == pod_ip
            and backend[0]["port_value"] == pod_port
            and backend[0]["eds_health_status"] == "HEALTHY"
        ]

        if not healthy_backends:
            return False

    return True


def are_namespaces_up_in_eds(
    envoy_eds_path: str,
    namespaces: Collection[str],
    pod_ip: str,
    pod_port: int,
) -> bool:
    """Returns whether a Pod is registered on Envoy through the EDS
    :param envoy_eds_path: path where EDS yaml files are stored
    :param namespaces: list of namespaces to check
    :param pod_ip: IP of the pod
    :param pod_port: The port to reach the service in the pod
    """

    for namespace in namespaces:
        backends_from_eds = get_backends_from_eds(namespace, envoy_eds_path)
        if (pod_ip, pod_port) not in backends_from_eds:
            return False

    return True


def retrieve_envoy_clusters(
    envoy_host: str, envoy_admin_port: int, envoy_admin_endpoint_format: str
) -> Dict[str, Any]:
    envoy_uri = envoy_admin_endpoint_format.format(
        host=envoy_host, port=envoy_admin_port, endpoint="clusters?format=json"
    )

    # timeout after 3 seconds and retry 3 times
    envoy_admin_request = requests.Session()
    envoy_admin_request.headers.update({"User-Agent": get_user_agent()})
    envoy_admin_request.mount("http://", requests.adapters.HTTPAdapter(max_retries=3))
    envoy_admin_request.mount("https://", requests.adapters.HTTPAdapter(max_retries=3))
    envoy_admin_response = envoy_admin_request.get(envoy_uri, timeout=3)
    return envoy_admin_response.json()


def get_casper_endpoints(
    clusters_info: Mapping[str, Any]
) -> FrozenSet[Tuple[str, int]]:
    """Filters out and returns casper endpoints from Envoy clusters."""
    casper_endpoints: Set[Tuple[str, int]] = set()
    for cluster_status in clusters_info["cluster_statuses"]:
        if "host_statuses" in cluster_status:
            if cluster_status["name"].startswith("spectre.") and cluster_status[
                "name"
            ].endswith(".egress_cluster"):
                for host_status in cluster_status["host_statuses"]:
                    casper_endpoints.add(
                        (
                            host_status["address"]["socket_address"]["address"],
                            host_status["address"]["socket_address"]["port_value"],
                        )
                    )
    return frozenset(casper_endpoints)


def get_backends_from_eds(namespace: str, envoy_eds_path: str) -> List[Tuple[str, int]]:
    """Returns a list of backends for a given namespace. Casper backends are also returned (if present).

    :param namespace: return backends for this namespace
    :param envoy_eds_path: path where EDS yaml files are stored
    :returns backends: a list of touples representing the backends for
                       the requested service
    """
    backends = []
    eds_file_for_namespace = f"{envoy_eds_path}/{namespace}/{namespace}.yaml"

    if os.access(eds_file_for_namespace, os.R_OK):
        with open(eds_file_for_namespace) as f:
            eds_yaml = yaml.safe_load(f)
            for resource in eds_yaml.get("resources", []):
                endpoints = resource.get("endpoints")
                # endpoints could be None if there are no backends listed
                if endpoints:
                    for endpoint in endpoints:
                        for lb_endpoint in endpoint.get("lb_endpoints", []):
                            address = lb_endpoint["endpoint"]["address"][
                                "socket_address"
                            ]["address"]
                            port_value = lb_endpoint["endpoint"]["address"][
                                "socket_address"
                            ]["port_value"]
                            backends.append((address, port_value))
    return backends


def get_backends(
    service: str,
    envoy_host: str,
    envoy_admin_port: int,
    envoy_admin_endpoint_format: str,
) -> Dict[str, List[Tuple[EnvoyBackend, bool]]]:
    """Fetches JSON from Envoy admin's /clusters endpoint and returns a list of backends.

    :param service: If None, return backends for all services, otherwise only return backends for this particular
                    service.
    :param envoy_host: The host that this check should contact for replication information.
    :param envoy_admin_port: The port that Envoy's admin interface is listening on
    :param envoy_admin_endpoint_format: The format of Envoy's admin endpoint
    :returns backends: A list of dicts representing the backends of all
                       services or the requested service
    """
    if service:
        services = [service]
    else:
        services = None
    return get_multiple_backends(
        services,
        envoy_host=envoy_host,
        envoy_admin_port=envoy_admin_port,
        envoy_admin_endpoint_format=envoy_admin_endpoint_format,
    )


def get_multiple_backends(
    services: Optional[Sequence[str]],
    envoy_host: str,
    envoy_admin_port: int,
    envoy_admin_endpoint_format: str,
    resolve_hostnames: bool = True,
) -> Dict[str, List[Tuple[EnvoyBackend, bool]]]:
    """Fetches JSON from Envoy admin's /clusters endpoint and returns a list of backends.

    :param services: If None, return backends for all services, otherwise only return backends for these particular
                     services.
    :param envoy_host: The host that this check should contact for replication information.
    :param envoy_admin_port: The port that Envoy's admin interface is listening on
    :param envoy_admin_endpoint_format: The format of Envoy's admin endpoint
    :returns backends: A list of dicts representing the backends of all
                       services or the requested service
    """
    clusters_info = retrieve_envoy_clusters(
        envoy_host=envoy_host,
        envoy_admin_port=envoy_admin_port,
        envoy_admin_endpoint_format=envoy_admin_endpoint_format,
    )

    casper_endpoints = get_casper_endpoints(clusters_info)

    backends: DefaultDict[
        str, List[Tuple[EnvoyBackend, bool]]
    ] = collections.defaultdict(list)
    for cluster_status in clusters_info["cluster_statuses"]:
        if "host_statuses" in cluster_status:
            if cluster_status["name"].endswith(".egress_cluster"):
                service_name = cluster_status["name"][: -len(".egress_cluster")]

                if services is None or service_name in services:
                    cluster_backends = []
                    casper_endpoint_found = False
                    for host_status in cluster_status["host_statuses"]:
                        address = host_status["address"]["socket_address"]["address"]
                        port_value = host_status["address"]["socket_address"][
                            "port_value"
                        ]

                        # Check if this endpoint is actually a casper backend
                        # If so, omit from the service's list of backends
                        if not service_name.startswith("spectre."):
                            if (address, port_value) in casper_endpoints:
                                casper_endpoint_found = True
                                continue

                        hostname = address
                        if resolve_hostnames:
                            try:
                                hostname = socket.gethostbyaddr(address)[0].split(".")[
                                    0
                                ]
                            except socket.herror:
                                # Default to the raw IP address if we can't lookup the hostname
                                pass

                        cluster_backends.append(
                            (
                                EnvoyBackend(
                                    address=address,
                                    port_value=port_value,
                                    hostname=hostname,
                                    eds_health_status=host_status["health_status"][
                                        "eds_health_status"
                                    ],
                                    weight=host_status["weight"],
                                ),
                                casper_endpoint_found,
                            )
                        )
                    backends[service_name] += cluster_backends
    return backends


def match_backends_and_pods(
    backends: Iterable[EnvoyBackend],
    pods: Iterable[V1Pod],
) -> List[Tuple[Optional[EnvoyBackend], Optional[V1Pod]]]:
    """Returns tuples of matching (backend, pod) pairs, as matched by IP. Each backend will be listed exactly
    once. If a backend does not match with a pod, (backend, None) will be included.
    If a pod's IP does not match with any backends, (None, pod) will be included.

    :param backends: An iterable of Envoy backend dictionaries, e.g. the list returned by
                     envoy_tools.get_multiple_backends.
    :param pods: A list of pods
    """

    # { ip : [backend1, backend2], ... }
    backends_by_ip: DefaultDict[str, List[EnvoyBackend]] = collections.defaultdict(list)
    backend_pod_pairs = []

    for backend in backends:
        ip = backend["address"]
        backends_by_ip[ip].append(backend)

    for pod in pods:
        ip = pod.status.pod_ip
        for backend in backends_by_ip.pop(ip, [None]):
            backend_pod_pairs.append((backend, pod))

    # we've been popping in the above loop, so anything left didn't match a k8s pod.
    for backends in backends_by_ip.values():
        for backend in backends:
            backend_pod_pairs.append((backend, None))

    return backend_pod_pairs


def build_envoy_location_dict(
    location: str,
    matched_envoy_backends_and_tasks: Sequence[
        Tuple[Optional[EnvoyBackend], Optional[V1Pod]]
    ],
    should_return_individual_backends: bool,
    casper_proxied_backends: AbstractSet[Tuple[str, int]],
) -> MutableMapping[str, Any]:
    running_backends_count = 0
    envoy_backends = []
    is_proxied_through_casper = False
    for backend, task in matched_envoy_backends_and_tasks:
        if backend is None:
            continue
        if backend["eds_health_status"] == "HEALTHY":
            running_backends_count += 1
        if should_return_individual_backends:
            backend["has_associated_task"] = task is not None
            envoy_backends.append(backend)
        if (backend["address"], backend["port_value"]) in casper_proxied_backends:
            is_proxied_through_casper = True
    return {
        "name": location,
        "running_backends_count": running_backends_count,
        "backends": envoy_backends,
        "is_proxied_through_casper": is_proxied_through_casper,
    }


def get_replication_for_all_services(
    envoy_host: str,
    envoy_admin_port: int,
    envoy_admin_endpoint_format: str,
) -> Dict[str, int]:
    """Returns the replication level for all services known to this Envoy

    :param envoy_host: The host that this check should contact for replication information.
    :param envoy_admin_port: The port number that this check should contact for replication information.
    :param envoy_admin_endpoint_format: The format of Envoy's admin endpoint
    :returns available_instance_counts: A dictionary mapping the service names
                                        to an integer number of available replicas.
    """
    backends = get_multiple_backends(
        services=None,
        envoy_host=envoy_host,
        envoy_admin_port=envoy_admin_port,
        envoy_admin_endpoint_format=envoy_admin_endpoint_format,
        resolve_hostnames=False,  # we're not really going to use the hostnames, so skip fetching them to save time
    )
    return collections.Counter(
        [
            service_name
            for service_name, service_backends in backends.items()
            for b in service_backends
            if backend_is_up(b[0])
        ]
    )


def backend_is_up(backend: EnvoyBackend) -> bool:
    return backend["eds_health_status"] == "HEALTHY"



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./setup_kubernetes_job.py <service.instance> [options]

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- -v, --verbose: Verbose output
"""
import argparse
import json
import logging
import sys
import time
import traceback
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from paasta_tools.eks_tools import EksDeploymentConfig
from paasta_tools.eks_tools import load_eks_service_config_no_cache
from paasta_tools.kubernetes.application.controller_wrappers import Application
from paasta_tools.kubernetes.application.controller_wrappers import (
    get_application_wrapper,
)
from paasta_tools.kubernetes_tools import AUTOSCALING_OVERRIDES_CONFIGMAP_NAME
from paasta_tools.kubernetes_tools import AUTOSCALING_OVERRIDES_CONFIGMAP_NAMESPACE
from paasta_tools.kubernetes_tools import ensure_namespace
from paasta_tools.kubernetes_tools import get_namespaced_configmap
from paasta_tools.kubernetes_tools import HpaOverride
from paasta_tools.kubernetes_tools import InvalidKubernetesConfig
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.kubernetes_tools import list_all_paasta_deployments
from paasta_tools.kubernetes_tools import load_kubernetes_service_config_no_cache
from paasta_tools.metrics import metrics_lib
from paasta_tools.utils import decompose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import InvalidJobNameError
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import NoConfigurationForServiceError
from paasta_tools.utils import NoDeploymentsAvailable
from paasta_tools.utils import SPACER

log = logging.getLogger(__name__)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Creates Kubernetes jobs.")
    parser.add_argument(
        "service_instance_list",
        nargs="+",
        help="The list of Kubernetes service instances to create or update",
        metavar="SERVICE%sINSTANCE" % SPACER,
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        "-c",
        "--cluster",
        dest="cluster",
        help="paasta cluster",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
    )
    parser.add_argument(
        "-l",
        "--rate-limit",
        dest="rate_limit",
        default=0,
        metavar="LIMIT",
        type=int,
        help="Update or create up to this number of service instances. Default is 0 (no limit).",
    )
    parser.add_argument(
        "--eks",
        help="This flag deploys only k8 services that should run on EKS",
        dest="eks",
        action="store_true",
        default=False,
    )
    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()
    soa_dir = args.soa_dir
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        # filter out unwanted zookeeper messages in the log
        logging.getLogger("kazoo").setLevel(logging.WARN)
        logging.basicConfig(level=logging.INFO)

    # emit deploy events for updated jobs
    deploy_metrics = metrics_lib.get_metrics_interface("paasta")

    # emit timing metrics for s_k_j
    cluster = args.cluster or load_system_paasta_config().get_cluster()
    timer = metrics_lib.system_timer(
        dimensions=dict(
            cluster=cluster,
            eks=args.eks,
        ),
    )
    timer.start()

    kube_client = KubeClient()
    service_instances_valid = True

    hpa_overrides = get_hpa_overrides(kube_client)

    # validate the service_instance names
    service_instances_with_valid_names = get_service_instances_with_valid_names(
        service_instances=args.service_instance_list
    )

    # returns a list of pairs of (No error?, KubernetesDeploymentConfig | EksDeploymentConfig) for every service_instance
    service_instance_configs_list = get_kubernetes_deployment_config(
        service_instances_with_valid_names=service_instances_with_valid_names,
        cluster=cluster,
        soa_dir=soa_dir,
        eks=args.eks,
    )

    if ((False, None) in service_instance_configs_list) or (
        len(service_instances_with_valid_names) != len(args.service_instance_list)
    ):
        service_instances_valid = False

    if service_instance_configs_list:
        for _, service_instance_config in service_instance_configs_list:
            if service_instance_config:
                ensure_namespace(
                    kube_client, namespace=service_instance_config.get_namespace()
                )

        setup_kube_succeeded = setup_kube_deployments(
            kube_client=kube_client,
            cluster=args.cluster or load_system_paasta_config().get_cluster(),
            service_instance_configs_list=service_instance_configs_list,
            rate_limit=args.rate_limit,
            soa_dir=soa_dir,
            metrics_interface=deploy_metrics,
            eks=args.eks,
            hpa_overrides=hpa_overrides,
        )
    else:
        setup_kube_succeeded = False
    exit_code = 0 if setup_kube_succeeded and service_instances_valid else 1

    timer.stop(tmp_dimensions={"result": exit_code})
    logging.info(
        f"Stopping timer for {cluster} (eks={args.eks}) with result {exit_code}: {timer()}ms elapsed"
    )
    sys.exit(exit_code)


def get_service_instances_with_valid_names(
    service_instances: Sequence[str],
) -> List[Tuple[str, str, str, str]]:
    service_instances_with_valid_names = [
        decompose_job_id(service_instance)
        for service_instance in service_instances
        if validate_job_name(service_instance)
    ]
    return service_instances_with_valid_names


def validate_job_name(service_instance: str) -> bool:
    try:
        service, instance, _, __ = decompose_job_id(service_instance)
    except InvalidJobNameError:
        log.error(
            "Invalid service instance specified. Format is service%sinstance." % SPACER
        )
        return False
    return True


def get_kubernetes_deployment_config(
    service_instances_with_valid_names: list,
    cluster: str,
    soa_dir: str = DEFAULT_SOA_DIR,
    eks: bool = False,
) -> List[Tuple[bool, Union[KubernetesDeploymentConfig, EksDeploymentConfig]]]:
    service_instance_configs_list = []
    for service_instance in service_instances_with_valid_names:
        try:
            service_instance_config: Union[
                KubernetesDeploymentConfig, EksDeploymentConfig
            ]
            if eks:
                service_instance_config = load_eks_service_config_no_cache(
                    service=service_instance[0],
                    instance=service_instance[1],
                    cluster=cluster,
                    soa_dir=soa_dir,
                )
            else:
                service_instance_config = load_kubernetes_service_config_no_cache(
                    service=service_instance[0],
                    instance=service_instance[1],
                    cluster=cluster,
                    soa_dir=soa_dir,
                )
            service_instance_configs_list.append((True, service_instance_config))
        except NoDeploymentsAvailable:
            log.debug(
                "No deployments found for %s.%s in cluster %s. Skipping."
                % (service_instance[0], service_instance[1], cluster)
            )
            service_instance_configs_list.append((True, None))
        except NoConfigurationForServiceError:
            error_msg = (
                f"Could not read kubernetes configuration file for %s.%s in cluster %s"
                % (service_instance[0], service_instance[1], cluster)
            )
            log.error(error_msg)
            service_instance_configs_list.append((False, None))
    return service_instance_configs_list


def get_hpa_overrides(kube_client: KubeClient) -> Dict[str, Dict[str, HpaOverride]]:
    """
    Load autoscaling overrides from the ConfigMap once.

    This function reads the "paasta-autoscaling-overrides" ConfigMap in the "paasta" namespace
    and extracts all valid (non-expired) overrides to return a dictionary mapping
    service.instance pairs to override data (currently, just min_instances and when the
    override should expire by).

    The incoming ConfigMap is expected to have the following format:
    {
        $SERVICE_A.$INSTANCE_A: {
            "min_instances": 2,
            "expire_after": "2023-10-01T00:00:00Z"
        },
        $SERVICE_A.$INSTANCE_B: {
            "min_instances": 3,
            "expire_after": "2023-10-01T00:00:00Z"
        },
            ...
        },
        $SERVICE_B.$INSTANCE_A: {
            "min_instances": 1,
            "expire_after": "2023-10-01T00:00:00Z"
        },
        $SERVICE_B.$INSTANCE_B: {
            "min_instances": 2,
            "expire_after": "2023-10-01T00:00:00Z"
        },
        ...
    }
    """
    overrides: Dict[str, Dict[str, HpaOverride]] = {}
    try:
        configmap = get_namespaced_configmap(
            name=AUTOSCALING_OVERRIDES_CONFIGMAP_NAME,
            namespace=AUTOSCALING_OVERRIDES_CONFIGMAP_NAMESPACE,
            kube_client=kube_client,
        )

        if configmap.data:
            current_time = time.time()

            for service_instance, override_json in configmap.data.items():
                try:
                    service, instance = service_instance.split(".")
                    override_metadata = json.loads(override_json)
                    expire_after = override_metadata.get("expire_after")
                    min_instances = override_metadata.get("min_instances")

                    if expire_after and min_instances:
                        if current_time < expire_after:
                            if service not in overrides:
                                overrides[service] = {}

                            overrides[service][instance] = {
                                "min_instances": min_instances,
                                "expire_after": expire_after,
                            }
                            log.info(
                                f"Found valid HPA override for {service}: "
                                f"{override_metadata.get('min_instances')} (expires at {expire_after})"
                            )
                        else:
                            log.info(
                                f"Ignoring expired HPA override for {service}.{instance}"
                                f"(expired at {expire_after})"
                            )
                    else:
                        log.warning(
                            f"Invalid HPA override for {service}.{instance}: "
                            f"missing 'min_instances' or 'expire_after': {override_metadata}"
                        )
                except Exception:
                    log.exception(
                        f"Error parsing override for {service} - proceeding without overrides for this service."
                    )
    except Exception:
        # If ConfigMap doesn't exist or there's an error, just return empty dict
        log.exception(
            f"Unable to load the {AUTOSCALING_OVERRIDES_CONFIGMAP_NAME} ConfigMap - proceeding without overrides"
        )

    return overrides


def setup_kube_deployments(
    kube_client: KubeClient,
    cluster: str,
    service_instance_configs_list: List[
        Tuple[bool, Union[KubernetesDeploymentConfig, EksDeploymentConfig]]
    ],
    rate_limit: int = 0,
    soa_dir: str = DEFAULT_SOA_DIR,
    metrics_interface: metrics_lib.BaseMetrics = metrics_lib.NoMetrics("paasta"),
    eks: bool = False,
    hpa_overrides: Optional[Dict[str, Dict[str, HpaOverride]]] = None,
) -> bool:

    if not service_instance_configs_list:
        return True

    existing_kube_deployments = set(list_all_paasta_deployments(kube_client))
    existing_apps = {
        (deployment.service, deployment.instance, deployment.namespace)
        for deployment in existing_kube_deployments
    }

    hpa_overrides = hpa_overrides or {}

    applications = [
        create_application_object(
            cluster=cluster,
            soa_dir=soa_dir,
            service_instance_config=service_instance,
            eks=eks,
            hpa_override=hpa_overrides.get(service_instance.service, {}).get(
                service_instance.instance, None
            ),
        )
        if service_instance
        else (_, None)
        for _, service_instance in service_instance_configs_list
    ]
    api_updates = 0
    for _, app in applications:
        if app:
            app_dimensions = {
                "paasta_service": app.kube_deployment.service,
                "paasta_instance": app.kube_deployment.instance,
                "paasta_cluster": cluster,
                "paasta_namespace": app.kube_deployment.namespace,
            }
            try:
                app.update_dependency_api_objects(kube_client)
                if (
                    app.kube_deployment.service,
                    app.kube_deployment.instance,
                    app.kube_deployment.namespace,
                ) not in existing_apps:
                    if app.soa_config.get_bounce_method() == "downthenup":
                        if any(
                            (
                                existing_app[:2]
                                == (
                                    app.kube_deployment.service,
                                    app.kube_deployment.instance,
                                )
                            )
                            for existing_app in existing_apps
                        ):
                            # For downthenup, we don't want to create until cleanup_kubernetes_job has cleaned up the instance in the other namespace.
                            continue
                    log.info(f"Creating {app} because it does not exist yet.")
                    app.create(kube_client)
                    app_dimensions["deploy_event"] = "create"
                    metrics_interface.emit_event(
                        name="deploy",
                        dimensions=app_dimensions,
                    )
                    api_updates += 1
                elif app.kube_deployment not in existing_kube_deployments:
                    log.info(f"Updating {app} because configs have changed.")
                    app.update(kube_client)
                    app_dimensions["deploy_event"] = "update"
                    metrics_interface.emit_event(
                        name="deploy",
                        dimensions=app_dimensions,
                    )
                    api_updates += 1
                else:
                    log.info(f"{app} is up-to-date!")

                log.info(f"Ensuring related API objects for {app} are in sync")
                app.update_related_api_objects(kube_client)
            except Exception:
                log.exception(f"Error while processing: {app}")
        if rate_limit > 0 and api_updates >= rate_limit:
            log.info(
                f"Not doing any further updates as we reached the limit ({api_updates})"
            )
            break
    return (False, None) not in applications


def create_application_object(
    cluster: str,
    soa_dir: str,
    service_instance_config: Union[KubernetesDeploymentConfig, EksDeploymentConfig],
    eks: bool = False,
    hpa_override: Optional[HpaOverride] = None,
) -> Tuple[bool, Optional[Application]]:
    try:
        formatted_application = service_instance_config.format_kubernetes_app()
    except InvalidKubernetesConfig:
        log.error(traceback.format_exc())
        return False, None

    app = get_application_wrapper(formatted_application, hpa_override)
    app.load_local_config(soa_dir, cluster, eks)

    return True, app


if __name__ == "__main__":
    main()



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest.mock import patch

import pytest

from paasta_tools import generate_authenticating_services
from paasta_tools import yaml_tools as yaml


@pytest.fixture
def mock_soa_configs(tmpdir):
    soa_dir = tmpdir / "soa_configs"
    soa_dir.mkdir()
    service_a = soa_dir / "service_a"
    service_a.mkdir()
    with (service_a / "authorization.yaml").open("w") as f:
        yaml.safe_dump(
            {
                "authorization": {
                    "rules": [
                        {"identity_groups": {"services": ["service_b", "service_c"]}},
                        {"identity_groups": {"services": ["service_d"]}},
                    ],
                },
            },
            f,
        ),
    service_b = soa_dir / "service_b"
    service_b.mkdir()
    with (service_b / "authorization.yaml").open("w") as f:
        yaml.safe_dump(
            {
                "authorization": {
                    "rules": [
                        {"identity_groups": {"services": ["service_a", "service_d"]}},
                    ],
                },
            },
            f,
        ),
    with patch(
        "paasta_tools.generate_authenticating_services.DEFAULT_SOA_DIR", str(soa_dir)
    ):
        yield


@pytest.fixture
def mock_paasta_config():
    with patch(
        "paasta_tools.generate_authenticating_services.load_system_paasta_config"
    ) as mock_load:
        mock_load.return_value.get_always_authenticating_services.return_value = [
            "service_always"
        ]
        yield mock_load.return_value


def test_enumerate_authenticating_services(mock_soa_configs, mock_paasta_config):
    assert generate_authenticating_services.enumerate_authenticating_services() == {
        "services": [
            "service_a",
            "service_always",
            "service_b",
            "service_c",
            "service_d",
        ],
    }


@patch("paasta_tools.utils.socket")
@patch("paasta_tools.utils.datetime")
@patch("paasta_tools.generate_authenticating_services.parse_args")
def test_main_yaml_config(
    mock_parse_args,
    mock_datetime,
    mock_socket,
    tmpdir,
    mock_soa_configs,
    mock_paasta_config,
):
    output = tmpdir / "authenticating.yaml"
    mock_datetime.datetime.now().isoformat.return_value = "$SOME_TIME"
    mock_socket.getfqdn.return_value = "somehost.yelp"
    mock_parse_args.return_value.output_filename = str(output)
    mock_parse_args.return_value.output_format = "yaml"

    generate_authenticating_services.main()

    with output.open() as f:
        assert (
            f.read()
            == """
# This file is automatically generated by paasta_tools.
# It was automatically generated at $SOME_TIME on somehost.yelp.
---
services:
- service_a
- service_always
- service_b
- service_c
- service_d
""".lstrip()
        )



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock

from paasta_tools import remote_git


def test_make_determine_wants_func():
    refs = {b"refs/heads/foo": b"abcde", b"refs/tags/blah": b"12345"}
    # nothing changed, so nothing should change
    determine_wants = remote_git._make_determine_wants_func(lambda x: x)
    assert determine_wants(refs) == refs

    # don't delete anything.
    determine_wants = remote_git._make_determine_wants_func(lambda x: {})
    assert determine_wants(refs) == refs

    # don't modify anything existing.
    determine_wants = remote_git._make_determine_wants_func(
        lambda x: {k: v[::-1] for k, v in x.items()}
    )
    assert determine_wants(refs) == refs

    # only allow new things
    determine_wants = remote_git._make_determine_wants_func(lambda x: {"foo": "bar"})
    actual = determine_wants(refs)
    expected = dict(refs)
    expected.update({b"foo": b"bar"})
    assert actual == expected


def test_non_ascii_tags():
    """git tags can be UTF-8 encoded"""

    with mock.patch(
        "dulwich.client.get_transport_and_path",
        autospec=True,
        return_value=(
            mock.Mock(
                **{"fetch_pack.return_value": {"☃".encode("UTF-8"): b"deadbeef"}}
            ),
            "path",
        ),
    ):
        with mock.patch("time.sleep", autospec=True):
            ret = remote_git.list_remote_refs("git-url")
            assert ret == {"☃": "deadbeef"}


def test_make_force_push_mutate_refs_func_overwrites_shas():
    targets = ["refs/heads/targeta", "refs/tags/targetb"]
    input_refs = {
        b"refs/heads/foo": b"12345",
        b"refs/heads/targeta": b"12345",
        b"refs/tags/targetb": b"12345",
        b"refs/heads/ignored": b"12345",
        b"refs/tags/blah": b"12345",
    }
    expected = {
        b"refs/heads/foo": b"12345",
        b"refs/heads/targeta": b"newsha",
        b"refs/tags/targetb": b"newsha",
        b"refs/heads/ignored": b"12345",
        b"refs/tags/blah": b"12345",
    }

    mutate_refs_func = remote_git.make_force_push_mutate_refs_func(
        targets=targets, sha="newsha"
    )
    actual = mutate_refs_func(input_refs)
    assert actual == expected
    assert all([isinstance(k, bytes) for k in actual])


@mock.patch("dulwich.client", autospec=True)
@mock.patch("paasta_tools.remote_git._make_determine_wants_func", autospec=True)
def test_create_remote_refs_is_safe_by_default(
    mock_make_determine_wants_func, mock_dulwich_client
):
    git_url = "fake_git_url"
    fake_git_client = mock.Mock()
    mock_dulwich_client.get_transport_and_path.return_value = (
        fake_git_client,
        "fake_path",
    )
    remote_git.create_remote_refs(
        git_url=git_url, ref_mutator=mock.sentinel.ref_mutator
    )
    fake_git_client.send_pack.assert_called_once_with(
        "fake_path", mock_make_determine_wants_func.return_value, mock.ANY
    )


@mock.patch("dulwich.client", autospec=True)
def test_create_remote_refs_allows_force_and_uses_the_provided_mutator(
    mock_dulwich_client,
):
    git_url = "fake_git_url"
    fake_git_client = mock.Mock()
    mock_dulwich_client.get_transport_and_path.return_value = (
        fake_git_client,
        "fake_path",
    )
    remote_git.create_remote_refs(
        git_url=git_url, ref_mutator=mock.sentinel.ref_mutator, force=True
    )
    fake_git_client.send_pack.assert_called_once_with(
        "fake_path", mock.sentinel.ref_mutator, mock.ANY
    )


@mock.patch("paasta_tools.remote_git._run", autospec=True)
def test_get_authors_fails_with_bad_url(mock_run):
    expected = 1, mock.ANY
    assert expected == remote_git.get_authors("bad", "a", "b")


@mock.patch("paasta_tools.remote_git._run", autospec=True)
def test_get_authors_fails_with_unknown(mock_run):
    url = "git@bitbucket.org:something.git"
    expected = 1, mock.ANY
    assert expected == remote_git.get_authors(url, "a", "b")


@mock.patch("paasta_tools.remote_git._run", autospec=True)
def test_get_authors_works_with_good_url(mock_run):
    mock_run.return_value = (0, "it worked")
    expected = mock_run.return_value
    assert expected == remote_git.get_authors(
        "git@git.yelpcorp.com:yelp-main", "a", "b"
    )
    mock_run.assert_called_once_with(
        command="ssh git@git.yelpcorp.com authors-of-changeset yelp-main a b",
        timeout=5.0,
    )



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock

from paasta_tools import bounce_lib
from paasta_tools import utils


class TestBounceLib:
    def fake_system_paasta_config(self):
        return utils.SystemPaastaConfig({"synapse_port": 123456}, "/fake/configs")

    def test_get_bounce_method_func(self):
        actual = bounce_lib.get_bounce_method_func("brutal")
        expected = bounce_lib.brutal_bounce
        assert actual == expected


class TestBrutalBounce:
    def test_brutal_bounce_no_existing_apps(self):
        """When marathon is unaware of a service, brutal bounce should try to
        create a marathon app."""
        new_config = {"id": "foo.bar.12345"}
        happy_tasks = []

        assert bounce_lib.brutal_bounce(
            new_config=new_config,
            new_app_running=False,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=[],
        ) == {"create_app": True, "tasks_to_drain": set()}

    def test_brutal_bounce_done(self):
        """When marathon has the desired app, and there are no other copies of
        the service running, brutal bounce should neither start nor stop
        anything."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(5)]

        assert bounce_lib.brutal_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=[],
        ) == {"create_app": False, "tasks_to_drain": set()}

    def test_brutal_bounce_mid_bounce(self):
        """When marathon has the desired app, but there are other copies of
        the service running, brutal bounce should stop the old ones."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(5)]

        assert bounce_lib.brutal_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {
            "create_app": False,
            "tasks_to_drain": set(
                old_app_live_happy_tasks + old_app_live_unhappy_tasks
            ),
        }

    def test_brutal_bounce_old_but_no_new(self):
        """When marathon does not have the desired app, but there are other copies of
        the service running, brutal bounce should stop the old ones and start
        the new one."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(5)]

        assert bounce_lib.brutal_bounce(
            new_config=new_config,
            new_app_running=False,
            happy_new_tasks=[],
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {
            "create_app": True,
            "tasks_to_drain": set(
                old_app_live_happy_tasks + old_app_live_unhappy_tasks
            ),
        }


class TestUpthendownBounce:
    def test_upthendown_bounce_no_existing_apps(self):
        """When marathon is unaware of a service, upthendown bounce should try to
        create a marathon app."""
        new_config = {"id": "foo.bar.12345"}
        happy_tasks = []

        assert bounce_lib.upthendown_bounce(
            new_config=new_config,
            new_app_running=False,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=[],
        ) == {"create_app": True, "tasks_to_drain": set()}

    def test_upthendown_bounce_old_but_no_new(self):
        """When marathon has the desired app, but there are other copies of
        the service running, upthendown bounce should start the new one. but
        not stop the old one yet."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(5)]

        assert bounce_lib.upthendown_bounce(
            new_config=new_config,
            new_app_running=False,
            happy_new_tasks=[],
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {"create_app": True, "tasks_to_drain": set()}

    def test_upthendown_bounce_mid_bounce(self):
        """When marathon has the desired app, and there are other copies of
        the service running, but the new app is not fully up, upthendown bounce
        should not stop the old ones."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(3)]
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(5)]

        assert bounce_lib.upthendown_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {"create_app": False, "tasks_to_drain": set()}

    def test_upthendown_bounce_cleanup(self):
        """When marathon has the desired app, and there are other copies of
        the service running, and the new app is fully up, upthendown bounce
        should stop the old ones."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(5)]

        assert bounce_lib.upthendown_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {
            "create_app": False,
            "tasks_to_drain": set(
                old_app_live_happy_tasks + old_app_live_unhappy_tasks
            ),
        }

    def test_upthendown_bounce_done(self):
        """When marathon has the desired app, and there are no other copies of
        the service running, upthendown bounce should neither start nor stop
        anything."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_happy_tasks = []
        old_app_live_unhappy_tasks = []

        assert bounce_lib.upthendown_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {"create_app": False, "tasks_to_drain": set()}


class TestCrossoverBounce:
    def test_crossover_bounce_no_existing_apps(self):
        """When marathon is unaware of a service, crossover bounce should try to
        create a marathon app."""
        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = []
        old_app_live_happy_tasks = []
        old_app_live_unhappy_tasks = []

        assert bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=False,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {"create_app": True, "tasks_to_drain": set()}

    def test_crossover_bounce_old_but_no_new(self):
        """When marathon only has old apps for this service, crossover bounce should start the new one, but not kill any
        old tasks yet."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = []
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = []

        assert bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=False,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {"create_app": True, "tasks_to_drain": set()}

    def test_crossover_bounce_old_app_is_happy_but_no_new_app_happy_tasks(self):
        """When marathon only has old apps for this service and margin_factor != 1,
        crossover bounce should start the new app and kill some old tasks."""

        new_config = {"id": "foo.bar.12345", "instances": 100}
        happy_tasks = []
        old_app_live_happy_tasks = [mock.Mock() for _ in range(100)]
        old_app_live_unhappy_tasks = []

        actual = bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=False,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
            margin_factor=0.95,
        )
        assert actual["create_app"] is True
        assert len(actual["tasks_to_drain"]) == 5

    def test_crossover_bounce_some_unhappy_old_some_happy_old_no_new(self):
        """When marathon only has old apps for this service, and some of them are unhappy (maybe they've been recently
        started), the crossover bounce should start a new app and prefer killing the unhappy tasks over the happy ones.
        """

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = []
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(5)]

        assert bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {"create_app": False, "tasks_to_drain": set(old_app_live_unhappy_tasks)}

    def test_crossover_bounce_some_unhappy_old_no_happy_old_no_new_tasks_no_excess(
        self,
    ):
        """When marathon only has old apps for this service, and all of their tasks are unhappy, and there are no excess
        tasks, the crossover bounce should start a new app and not kill any old tasks.
        """

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = []
        old_app_live_happy_tasks = []
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(5)]

        assert bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {"create_app": False, "tasks_to_drain": set()}

    def test_crossover_bounce_lots_of_unhappy_old_no_happy_old_no_new(self):
        """When marathon has a new app and multiple old apps, no new tasks are up, all old tasks are unhappy, and there
        are too many tasks running, the crossover bounce should kill some (but not all) of the old
        tasks.

        This represents a situation where
        """

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = []
        old_app_live_happy_tasks = []
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(10)]

        actual = bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        )
        assert actual["create_app"] is False
        assert len(actual["tasks_to_drain"]) == 5

    def test_crossover_bounce_lots_of_unhappy_old_some_happy_old_new_app_exists_no_new_tasks(
        self,
    ):
        """When marathon has a new app and multiple old apps, no new tasks are up, one of the old apps is healthy and
        the other is not, only unhealthy tasks should get killed.
        """

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = []
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(5)]

        actual = bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        )
        assert actual["create_app"] is False
        assert actual["tasks_to_drain"] == set(old_app_live_unhappy_tasks)
        # Since there are plenty of unhappy old tasks, we should not kill any new ones.
        assert len(actual["tasks_to_drain"] & set(old_app_live_happy_tasks)) == 0

    def test_crossover_bounce_mid_bounce(self):
        """When marathon has the desired app, and there are other copies of the service running, but the new app is not
        fully up, crossover bounce should only stop a few of the old instances."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(3)]
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = []

        actual = bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        )

        assert actual["create_app"] is False
        assert len(actual["tasks_to_drain"]) == 3

    def test_crossover_bounce_mid_bounce_some_happy_old_some_unhappy_old(self):
        """When marathon has the desired app, and there are other copies of the service running, and some of those
        older tasks are unhappy, we should prefer killing the unhappy tasks."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(3)]
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(1)]

        actual = bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        )

        assert actual["create_app"] is False
        assert len(actual["tasks_to_drain"]) == 4
        # There are fewer unhappy old tasks than excess tasks, so we should kill all unhappy old ones, plus a few
        # happy ones.
        assert set(old_app_live_unhappy_tasks).issubset(actual["tasks_to_drain"])

    def test_crossover_bounce_mid_bounce_some_happy_old_lots_of_unhappy_old(self):
        """When marathon has the desired app, and there are other copies of the service running, and there are more
        unhappy old tasks than excess tasks, we should only kill unhappy tasks.
        """

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(3)]
        old_app_live_happy_tasks = [mock.Mock() for _ in range(2)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(5)]

        actual = bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        )

        assert actual["create_app"] is False
        # There are as many unhappy old tasks as excess tasks, so all tasks that we kill should be old unhappy ones.
        assert len(actual["tasks_to_drain"]) == 5
        assert actual["tasks_to_drain"] == set(old_app_live_unhappy_tasks)

    def test_crossover_bounce_mid_bounce_no_happy_old_lots_of_unhappy_old(self):
        """When marathon has the desired app, and there are other copies of the service running, but none of the old
        tasks are happy, and there are excess tasks, we should kill some (but not all) unhappy old tasks."""
        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(3)]
        old_app_live_happy_tasks = []
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(6)]

        actual = bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        )
        assert actual["create_app"] is False
        assert len(actual["tasks_to_drain"]) == 4

    def test_crossover_bounce_using_margin_factor_big_numbers(self):
        new_config = {"id": "foo.bar.12345", "instances": 500}
        happy_tasks = [mock.Mock() for _ in range(100)]
        old_app_live_happy_tasks = [mock.Mock() for _ in range(300)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(100)]

        actual = bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
            margin_factor=0.95,
        )
        assert actual["create_app"] is False
        assert len(actual["tasks_to_drain"]) == 25

    def test_crossover_bounce_using_margin_factor_small_numbers(self):
        new_config = {"id": "foo.bar.12345", "instances": 3}
        happy_tasks = []
        old_app_live_happy_tasks = [mock.Mock() for _ in range(3)]
        old_app_live_unhappy_tasks = []

        actual = bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
            margin_factor=0.66,
        )
        assert actual["create_app"] is False
        assert len(actual["tasks_to_drain"]) == 1

    def test_crossover_bounce_done(self):
        """When marathon has the desired app, and there are no other copies of
        the service running, crossover bounce should neither start nor stop
        anything."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_happy_tasks = []
        old_app_live_unhappy_tasks = []

        assert bounce_lib.crossover_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {"create_app": False, "tasks_to_drain": set()}


class TestDownThenUpBounce:
    def test_downthenup_bounce_no_existing_apps(self):
        """When marathon is unaware of a service, downthenup bounce should try to
        create a marathon app."""
        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = []
        old_app_live_happy_tasks = []
        old_app_live_unhappy_tasks = []

        assert bounce_lib.downthenup_bounce(
            new_config=new_config,
            new_app_running=False,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {"create_app": True, "tasks_to_drain": set()}

    def test_downthenup_bounce_old_but_no_new(self):
        """When marathon has only old copies of the service, downthenup_bounce should kill them and not start a new one
        yet."""
        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = []
        old_app_live_happy_tasks = [mock.Mock() for _ in range(5)]
        old_app_live_unhappy_tasks = [mock.Mock() for _ in range(1)]

        assert bounce_lib.downthenup_bounce(
            new_config=new_config,
            new_app_running=False,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=old_app_live_happy_tasks
            + old_app_live_unhappy_tasks,
        ) == {
            "create_app": False,
            "tasks_to_drain": set(
                old_app_live_happy_tasks + old_app_live_unhappy_tasks
            ),
        }

    def test_downthenup_bounce_done(self):
        """When marathon has the desired app, and there are no other copies of the service running, downthenup bounce
        should neither start nor stop anything."""

        new_config = {"id": "foo.bar.12345", "instances": 5}
        happy_tasks = [mock.Mock() for _ in range(5)]

        assert bounce_lib.downthenup_bounce(
            new_config=new_config,
            new_app_running=True,
            happy_new_tasks=happy_tasks,
            old_non_draining_tasks=[],
        ) == {"create_app": False, "tasks_to_drain": set()}


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import docker
import mock
import pytest

from paasta_tools.paasta_execute_docker_command import execute_in_container
from paasta_tools.paasta_execute_docker_command import main
from paasta_tools.paasta_execute_docker_command import TimeoutException


def test_execute_in_container():
    fake_container_id = "fake_container_id"
    fake_return_code = 0
    fake_output = "fake_output"
    fake_command = "fake_cmd"
    mock_docker_client = mock.MagicMock(spec_set=docker.APIClient)
    mock_docker_client.exec_start.return_value = fake_output
    mock_docker_client.exec_inspect.return_value = {"ExitCode": fake_return_code}

    assert execute_in_container(
        mock_docker_client, fake_container_id, fake_command, 1
    ) == (fake_output, fake_return_code)
    expected_cmd = ["/bin/sh", "-c", fake_command]
    mock_docker_client.exec_create.assert_called_once_with(
        fake_container_id, expected_cmd
    )


@mock.patch(
    "paasta_tools.paasta_execute_docker_command.is_using_unprivileged_containers",
    lambda: False,
    autospec=None,
)
def test_execute_in_container_reuses_exec():
    fake_container_id = "fake_container_id"
    fake_execid = "fake_execid"
    fake_return_code = 0
    fake_output = "fake_output"
    fake_command = "fake_cmd"
    mock_docker_client = mock.MagicMock(spec_set=docker.APIClient)
    mock_docker_client.inspect_container.return_value = {"ExecIDs": [fake_execid]}
    mock_docker_client.exec_start.return_value = fake_output
    mock_docker_client.exec_inspect.return_value = {
        "ExitCode": fake_return_code,
        "ProcessConfig": {"entrypoint": "/bin/sh", "arguments": ["-c", fake_command]},
    }

    assert execute_in_container(
        mock_docker_client, fake_container_id, fake_command, 1
    ) == (fake_output, fake_return_code)
    assert mock_docker_client.exec_create.call_count == 0
    mock_docker_client.exec_start.assert_called_once_with(fake_execid, stream=False)


@mock.patch(
    "paasta_tools.paasta_execute_docker_command.is_using_unprivileged_containers",
    lambda: False,
    autospec=None,
)
def test_execute_in_container_reuses_only_valid_exec():
    fake_container_id = "fake_container_id"
    fake_execid = "fake_execid"
    fake_return_code = 0
    fake_output = "fake_output"
    fake_command = "fake_cmd"
    bad_exec = {
        "ExitCode": fake_return_code,
        "ProcessConfig": {
            "entrypoint": "/bin/sh",
            "arguments": ["-c", "some_other_command"],
        },
    }
    good_exec = {
        "ExitCode": fake_return_code,
        "ProcessConfig": {"entrypoint": "/bin/sh", "arguments": ["-c", fake_command]},
    }
    mock_docker_client = mock.MagicMock(spec_set=docker.APIClient)
    mock_docker_client.inspect_container.return_value = {
        "ExecIDs": ["fake_other_exec", fake_execid, "fake_other_exec"]
    }
    mock_docker_client.exec_start.return_value = fake_output
    # the last side effect is used to check the exit code of the command
    mock_docker_client.exec_inspect.side_effect = [
        bad_exec,
        good_exec,
        bad_exec,
        good_exec,
    ]

    assert execute_in_container(
        mock_docker_client, fake_container_id, fake_command, 1
    ) == (fake_output, fake_return_code)
    assert mock_docker_client.exec_create.call_count == 0
    mock_docker_client.exec_start.assert_called_once_with(fake_execid, stream=False)


def test_main():
    fake_container_id = "fake_container_id"
    fake_timeout = 3
    with mock.patch(
        "paasta_tools.paasta_execute_docker_command.get_container_id_for_mesos_id",
        return_value=fake_container_id,
        autospec=True,
    ), mock.patch(
        "paasta_tools.paasta_execute_docker_command.parse_args", autospec=True
    ) as args_patch, mock.patch(
        "paasta_tools.paasta_execute_docker_command.execute_in_container",
        return_value=("fake_output", 0),
        autospec=True,
    ), mock.patch(
        "paasta_tools.paasta_execute_docker_command.time_limit", autospec=True
    ) as time_limit_patch:
        args_patch.return_value.mesos_id = "fake_task_id"
        args_patch.return_value.timeout = fake_timeout
        with pytest.raises(SystemExit) as excinfo:
            main()
        time_limit_patch.assert_called_once_with(fake_timeout)
        assert excinfo.value.code == 0


def test_main_with_empty_task_id():
    fake_container_id = "fake_container_id"
    fake_timeout = 3
    with mock.patch(
        "paasta_tools.paasta_execute_docker_command.get_container_id_for_mesos_id",
        return_value=fake_container_id,
        autospec=True,
    ), mock.patch(
        "paasta_tools.paasta_execute_docker_command.parse_args", autospec=True
    ) as args_patch, mock.patch(
        "paasta_tools.paasta_execute_docker_command.execute_in_container",
        return_value=("fake_output", 0),
        autospec=True,
    ), mock.patch(
        "paasta_tools.paasta_execute_docker_command.time_limit", autospec=True
    ):
        args_patch.return_value.mesos_id = ""
        args_patch.return_value.timeout = fake_timeout
        with pytest.raises(SystemExit) as excinfo:
            main()
        assert excinfo.value.code == 2


def test_main_container_not_found_failure():
    with mock.patch(
        "paasta_tools.paasta_execute_docker_command.get_container_id_for_mesos_id",
        return_value=None,
        autospec=True,
    ), mock.patch(
        "paasta_tools.paasta_execute_docker_command.execute_in_container",
        return_value=("fake_output", 2),
        autospec=True,
    ), mock.patch(
        "paasta_tools.paasta_execute_docker_command.parse_args", autospec=True
    ) as args_patch, mock.patch(
        "paasta_tools.paasta_execute_docker_command.time_limit", autospec=True
    ):
        args_patch.return_value.mesos_id = "fake_task_id"
        with pytest.raises(SystemExit) as excinfo:
            main()
        assert excinfo.value.code == 1


def test_main_cmd_unclean_exit_failure():
    fake_container_id = "fake_container_id"
    with mock.patch(
        "paasta_tools.paasta_execute_docker_command.get_container_id_for_mesos_id",
        return_value=fake_container_id,
        autospec=True,
    ), mock.patch(
        "paasta_tools.paasta_execute_docker_command.execute_in_container",
        return_value=("fake_output", 2),
        autospec=True,
    ), mock.patch(
        "paasta_tools.paasta_execute_docker_command.parse_args", autospec=True
    ) as args_patch, mock.patch(
        "paasta_tools.paasta_execute_docker_command.time_limit", autospec=True
    ):
        args_patch.return_value.mesos_id = "fake_task_id"
        with pytest.raises(SystemExit) as excinfo:
            main()
        assert excinfo.value.code == 2


def test_main_timeout_failure():
    fake_container_id = "fake_container_id"
    fake_timeout = 3
    with mock.patch(
        "paasta_tools.paasta_execute_docker_command.get_container_id_for_mesos_id",
        return_value=fake_container_id,
        autospec=True,
    ), mock.patch(
        "paasta_tools.paasta_execute_docker_command.parse_args", autospec=True
    ) as args_patch, mock.patch(
        "paasta_tools.paasta_execute_docker_command.execute_in_container",
        return_value=("fake_output", 0),
        autospec=True,
    ), mock.patch(
        "paasta_tools.paasta_execute_docker_command.time_limit",
        side_effect=TimeoutException,
        autospec=True,
    ) as time_limit_patch:
        args_patch.return_value.mesos_id = "fake_task_id"
        args_patch.return_value.timeout = fake_timeout
        with pytest.raises(SystemExit) as excinfo:
            main()
        time_limit_patch.assert_called_once_with(fake_timeout)
        assert excinfo.value.code == 1



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import lru_cache
from typing import Optional

from paasta_tools.utils import load_system_paasta_config


try:
    from vault_tools.oidc import get_instance_oidc_identity_token
    from okta_auth import get_and_cache_jwt_default
except ImportError:

    def get_instance_oidc_identity_token(
        role: str, ecosystem: Optional[str] = None
    ) -> str:
        return ""

    def get_and_cache_jwt_default(client_id: str) -> str:
        return ""


@lru_cache(maxsize=1)
def get_service_auth_token() -> str:
    """Uses instance profile to authenticate with Vault and generate token for service authentication"""
    vault_role = load_system_paasta_config().get_service_auth_vault_role()
    return get_instance_oidc_identity_token(vault_role)


def get_sso_auth_token(paasta_apis: bool = False) -> str:
    """Generate an authentication token for the calling user from the Single Sign On provider

    :param bool paasta_apis: authenticate for PaaSTA APIs
    """
    system_config = load_system_paasta_config()
    client_id = (
        system_config.get_api_auth_sso_oidc_client_id()
        if paasta_apis
        else system_config.get_service_auth_sso_oidc_client_id()
    )
    return get_and_cache_jwt_default(client_id)



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
from pytest import raises

from paasta_tools import long_running_service_tools
from paasta_tools.utils import InvalidInstanceConfig


class TestLongRunningServiceConfig:
    def test_get_healthcheck_cmd_happy(self):
        fake_conf = long_running_service_tools.LongRunningServiceConfig(
            service="fake_name",
            cluster="fake_cluster",
            config_dict={"healthcheck_cmd": "/bin/true"},
            instance="fake_instance",
            branch_dict=None,
        )
        actual = fake_conf.get_healthcheck_cmd()
        assert actual == "/bin/true"

    def test_get_healthcheck_cmd_raises_when_unset(self):
        fake_conf = long_running_service_tools.LongRunningServiceConfig(
            service="fake_name",
            cluster="fake_cluster",
            instance="fake_instance",
            config_dict={},
            branch_dict=None,
        )
        with raises(InvalidInstanceConfig) as exc:
            fake_conf.get_healthcheck_cmd()
        assert "healthcheck mode 'cmd' requires a healthcheck_cmd to run" in str(
            exc.value
        )

    def test_get_healthcheck_for_instance_http(self):
        fake_service = "fake_service"
        fake_namespace = "fake_namespace"
        fake_hostname = "fake_hostname"
        fake_random_port = 666

        fake_path = "/fake_path"
        fake_service_config = long_running_service_tools.LongRunningServiceConfig(
            service=fake_service,
            cluster="fake_cluster",
            instance=fake_namespace,
            config_dict={},
            branch_dict=None,
        )
        fake_service_namespace_config = (
            long_running_service_tools.ServiceNamespaceConfig(
                {"mode": "http", "healthcheck_uri": fake_path}
            )
        )
        with mock.patch(
            "paasta_tools.long_running_service_tools.load_service_namespace_config",
            autospec=True,
            return_value=fake_service_namespace_config,
        ), mock.patch("socket.getfqdn", autospec=True, return_value=fake_hostname):
            expected = (
                "http",
                "http://%s:%d%s" % (fake_hostname, fake_random_port, fake_path),
            )
            actual = long_running_service_tools.get_healthcheck_for_instance(
                fake_service, fake_namespace, fake_service_config, fake_random_port
            )
            assert expected == actual

    def test_get_healthcheck_for_instance_not_matching_mode(self):
        fake_service = "fake_service"
        fake_namespace = "fake_namespace"
        fake_hostname = "fake_hostname"
        fake_random_port = 666

        fake_service_config = long_running_service_tools.LongRunningServiceConfig(
            service=fake_service,
            cluster="fake_cluster",
            instance=fake_namespace,
            config_dict={},
            branch_dict=None,
        )
        fake_service_namespace_config = (
            long_running_service_tools.ServiceNamespaceConfig({"mode": "http"})
        )
        with mock.patch(
            "paasta_tools.long_running_service_tools.load_service_namespace_config",
            autospec=True,
            return_value=fake_service_namespace_config,
        ), mock.patch("socket.getfqdn", autospec=True, return_value=fake_hostname):
            expected = ("http", "http://fake_hostname:666/status")
            actual = long_running_service_tools.get_healthcheck_for_instance(
                fake_service, fake_namespace, fake_service_config, fake_random_port
            )
            assert expected == actual

    def test_get_healthcheck_for_instance_tcp(self):
        fake_service = "fake_service"
        fake_namespace = "fake_namespace"
        fake_hostname = "fake_hostname"
        fake_random_port = 666

        fake_service_config = long_running_service_tools.LongRunningServiceConfig(
            service=fake_service,
            cluster="fake_cluster",
            instance=fake_namespace,
            config_dict={},
            branch_dict=None,
        )
        fake_service_namespace_config = (
            long_running_service_tools.ServiceNamespaceConfig({"mode": "tcp"})
        )
        with mock.patch(
            "paasta_tools.long_running_service_tools.load_service_namespace_config",
            autospec=True,
            return_value=fake_service_namespace_config,
        ), mock.patch("socket.getfqdn", autospec=True, return_value=fake_hostname):
            expected = ("tcp", "tcp://%s:%d" % (fake_hostname, fake_random_port))
            actual = long_running_service_tools.get_healthcheck_for_instance(
                fake_service, fake_namespace, fake_service_config, fake_random_port
            )
            assert expected == actual

    def test_get_healthcheck_for_instance_cmd(self):
        fake_service = "fake_service"
        fake_namespace = "fake_namespace"
        fake_hostname = "fake_hostname"
        fake_random_port = 666
        fake_cmd = "/bin/fake_command"
        fake_service_config = long_running_service_tools.LongRunningServiceConfig(
            service=fake_service,
            cluster="fake_cluster",
            instance=fake_namespace,
            config_dict={
                "instances": 1,
                "healthcheck_mode": "cmd",
                "healthcheck_cmd": fake_cmd,
            },
            branch_dict=None,
        )
        fake_service_namespace_config = (
            long_running_service_tools.ServiceNamespaceConfig({})
        )
        with mock.patch(
            "paasta_tools.long_running_service_tools.load_service_namespace_config",
            autospec=True,
            return_value=fake_service_namespace_config,
        ), mock.patch("socket.getfqdn", autospec=True, return_value=fake_hostname):
            expected = ("cmd", fake_cmd)
            actual = long_running_service_tools.get_healthcheck_for_instance(
                fake_service, fake_namespace, fake_service_config, fake_random_port
            )
            assert expected == actual

    def test_get_healthcheck_for_instance_other(self):
        fake_service = "fake_service"
        fake_namespace = "fake_namespace"
        fake_hostname = "fake_hostname"
        fake_random_port = 666
        fake_service_config = long_running_service_tools.LongRunningServiceConfig(
            service=fake_service,
            cluster="fake_cluster",
            instance=fake_namespace,
            config_dict={"healthcheck_mode": None},
            branch_dict=None,
        )
        fake_service_namespace_config = (
            long_running_service_tools.ServiceNamespaceConfig({})
        )
        with mock.patch(
            "paasta_tools.long_running_service_tools.load_service_namespace_config",
            autospec=True,
            return_value=fake_service_namespace_config,
        ), mock.patch("socket.getfqdn", autospec=True, return_value=fake_hostname):
            expected = (None, None)
            actual = long_running_service_tools.get_healthcheck_for_instance(
                fake_service, fake_namespace, fake_service_config, fake_random_port
            )
            assert expected == actual

    def test_get_healthcheck_for_instance_custom_soadir(self):
        fake_service = "fake_service"
        fake_namespace = "fake_namespace"
        fake_hostname = "fake_hostname"
        fake_random_port = 666
        fake_soadir = "/fake/soadir"
        fake_service_config = long_running_service_tools.LongRunningServiceConfig(
            service=fake_service,
            cluster="fake_cluster",
            instance=fake_namespace,
            config_dict={"healthcheck_mode": None},
            branch_dict=None,
        )
        fake_service_namespace_config = (
            long_running_service_tools.ServiceNamespaceConfig({})
        )
        with mock.patch(
            "paasta_tools.long_running_service_tools.load_service_namespace_config",
            autospec=True,
            return_value=fake_service_namespace_config,
        ) as load_service_namespace_config_patch, mock.patch(
            "socket.getfqdn", autospec=True, return_value=fake_hostname
        ):
            expected = (None, None)
            actual = long_running_service_tools.get_healthcheck_for_instance(
                fake_service,
                fake_namespace,
                fake_service_config,
                fake_random_port,
                soa_dir=fake_soadir,
            )
            assert expected == actual
            load_service_namespace_config_patch.assert_called_once_with(
                fake_service, fake_namespace, fake_soadir
            )

    def test_get_instances_in_config(self):
        fake_conf = long_running_service_tools.LongRunningServiceConfig(
            service="fake_name",
            cluster="fake_cluster",
            instance="fake_instance",
            config_dict={"instances": -10},
            branch_dict={
                "desired_state": "start",
                "git_sha": "c0ded00d",
                "docker_image": "docker_image",
                "image_version": None,
                "force_bounce": None,
            },
        )
        assert fake_conf.get_instances() == -10

    def test_get_instances_default(self):
        fake_conf = long_running_service_tools.LongRunningServiceConfig(
            service="fake_name",
            cluster="fake_cluster",
            instance="fake_instance",
            config_dict={},
            branch_dict=None,
        )
        assert fake_conf.get_instances() == 1

    def test_get_instances_respects_false(self):
        fake_conf = long_running_service_tools.LongRunningServiceConfig(
            service="fake_name",
            cluster="fake_cluster",
            instance="fake_instance",
            config_dict={"instances": False},
            branch_dict={
                "desired_state": "start",
                "git_sha": "c0debabe",
                "docker_image": "docker_image",
                "image_version": None,
                "force_bounce": None,
            },
        )
        assert fake_conf.get_instances() == 0

    def test_validate_with_bad_registration(self):
        fake_conf = long_running_service_tools.LongRunningServiceConfig(
            service="fake_name",
            cluster="fake_cluster",
            instance="fake_instance",
            config_dict={
                "registrations": ["fake_name.fake_instance", "bad_registration"],
                "deploy_group": None,
            },
            branch_dict=None,
        )
        error_messages = fake_conf.validate()
        assert "bad_registration" in error_messages[0]


class TestServiceNamespaceConfig:
    def test_get_mode_default(self):
        assert long_running_service_tools.ServiceNamespaceConfig().get_mode() is None

    def test_get_mode_default_when_port_specified(self):
        config = {"proxy_port": 1234}
        assert (
            long_running_service_tools.ServiceNamespaceConfig(config).get_mode()
            == "http"
        )

    def test_get_mode_valid(self):
        config = {"mode": "tcp"}
        assert (
            long_running_service_tools.ServiceNamespaceConfig(config).get_mode()
            == "tcp"
        )

    def test_get_mode_invalid(self):
        config = {"mode": "paasta"}
        with raises(long_running_service_tools.InvalidSmartstackMode):
            long_running_service_tools.ServiceNamespaceConfig(config).get_mode()

    def test_get_healthcheck_uri_default(self):
        assert (
            long_running_service_tools.ServiceNamespaceConfig().get_healthcheck_uri()
            == "/status"
        )

    def test_get_discover_default(self):
        assert (
            long_running_service_tools.ServiceNamespaceConfig().get_discover()
            == "region"
        )


def test_get_proxy_port_for_instance():
    mock_config = mock.Mock(
        get_registrations=mock.Mock(return_value=["thing.main.sha.sha"]),
        soa_dir="/nail/blah",
    )
    with mock.patch(
        "paasta_tools.long_running_service_tools.load_service_namespace_config",
        autospec=True,
    ) as mock_load_service_namespace_config:
        mock_load_service_namespace_config.return_value = {"proxy_port": 1234}
        assert (
            long_running_service_tools.get_proxy_port_for_instance(mock_config) == 1234
        )
        mock_load_service_namespace_config.assert_called_once_with(
            service="thing", namespace="main", soa_dir="/nail/blah"
        )


def test_host_passes_blacklist_passes():
    slave_attributes = {"fake_attribute": "fake_value_1"}
    blacklist = [("fake_attribute", "No what we have here"), ("foo", "bar")]
    actual = long_running_service_tools.host_passes_blacklist(
        host_attributes=slave_attributes, blacklist=blacklist
    )
    assert actual is True


def test_host_passes_blacklist_blocks_blacklisted_locations():
    slave_attributes = {"fake_attribute": "fake_value_1"}
    blacklist = [("fake_attribute", "fake_value_1")]
    actual = long_running_service_tools.host_passes_blacklist(
        host_attributes=slave_attributes, blacklist=blacklist
    )
    assert actual is False


def test_host_passes_whitelist():
    fake_slave_attributes = {
        "location_type": "fake_location",
        "fake_location_type": "fake_location",
    }
    fake_whitelist_allow = ("fake_location_type", ["fake_location"])
    fake_whitelist_deny = ("anoterfake_location_type", ["anotherfake_location"])

    slave_passes = long_running_service_tools.host_passes_whitelist(
        fake_slave_attributes, fake_whitelist_deny
    )
    assert not slave_passes
    slave_passes = long_running_service_tools.host_passes_whitelist(
        fake_slave_attributes, fake_whitelist_allow
    )
    assert slave_passes
    slave_passes = long_running_service_tools.host_passes_whitelist(
        fake_slave_attributes, None
    )
    assert slave_passes


def test_get_all_namespaces_for_service():
    name = "vvvvvv"
    soa_dir = "^_^"
    t1_dict = {"hollo": "werld", "smark": "stact"}
    t2_dict = {"vataman": "witir", "sin": "chaps"}
    fake_smartstack = {"smartstack": {"t1": t1_dict, "t2": t2_dict}}
    expected = [("vvvvvv.t2", t2_dict), ("vvvvvv.t1", t1_dict)]
    expected_short = [("t2", t2_dict), ("t1", t1_dict)]
    with mock.patch(
        "service_configuration_lib.read_service_configuration",
        autospec=True,
        return_value=fake_smartstack,
    ) as read_service_configuration_patch:
        actual = long_running_service_tools.get_all_namespaces_for_service(
            name, soa_dir
        )
        read_service_configuration_patch.assert_any_call(name, soa_dir)
        assert sorted(expected) == sorted(actual)

        actual_short = long_running_service_tools.get_all_namespaces_for_service(
            name, soa_dir, False
        )
        read_service_configuration_patch.assert_any_call(name, soa_dir)
        assert sorted(expected_short) == sorted(actual_short)


def test_get_all_namespaces():
    soa_dir = "carbon"
    namespaces = [
        [("aluminum", {"hydrogen": 1}), ("potassium", {"helium": 2})],
        [("uranium", {"lithium": 3}), ("gold", {"boron": 5})],
    ]
    expected = [
        ("uranium", {"lithium": 3}),
        ("gold", {"boron": 5}),
        ("aluminum", {"hydrogen": 1}),
        ("potassium", {"helium": 2}),
    ]
    with mock.patch(
        "os.path.abspath", autospec=True, return_value="oxygen"
    ) as abspath_patch, mock.patch(
        "os.listdir", autospec=True, return_value=["rid1", "rid2"]
    ) as listdir_patch, mock.patch(
        "paasta_tools.long_running_service_tools.get_all_namespaces_for_service",
        autospec=True,
        side_effect=lambda a, b: namespaces.pop(),
    ) as get_namespaces_patch:
        actual = long_running_service_tools.get_all_namespaces(soa_dir)
        assert expected == actual
        abspath_patch.assert_called_once_with(soa_dir)
        listdir_patch.assert_called_once_with("oxygen")
        get_namespaces_patch.assert_any_call("rid1", soa_dir)
        get_namespaces_patch.assert_any_call("rid2", soa_dir)
        assert get_namespaces_patch.call_count == 2


def test_get_expected_instance_count_for_namespace():
    service = "red"
    namespace = "rojo"
    soa_dir = "que_esta"
    fake_job_configs = [
        long_running_service_tools.LongRunningServiceConfig(
            service=service,
            cluster="fake_cluster",
            instance="blue",
            config_dict={"nerve_ns": "rojo", "instances": 11},
            branch_dict=None,
        ),
        long_running_service_tools.LongRunningServiceConfig(
            service=service,
            cluster="fake_cluster",
            instance="green",
            config_dict={"nerve_ns": "amarillo"},
            branch_dict=None,
        ),
    ]

    with mock.patch(
        "paasta_tools.long_running_service_tools.PaastaServiceConfigLoader",
        autospec=True,
        return_value=mock.Mock(
            instance_configs=mock.Mock(return_value=fake_job_configs)
        ),
    ) as fake_pscl:
        actual = long_running_service_tools.get_expected_instance_count_for_namespace(
            service,
            namespace,
            cluster="fake_cluster",
            soa_dir=soa_dir,
            instance_type_class=long_running_service_tools.LongRunningServiceConfig,
        )
        assert actual == 11
        fake_pscl.return_value.instance_configs.assert_called_once_with(
            cluster="fake_cluster",
            instance_type_class=long_running_service_tools.LongRunningServiceConfig,
        )



#!/usr/bin/env python
# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import logging
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import Any
from typing import Callable
from typing import Sequence
from typing import Tuple

from paasta_tools.kubernetes.remote_run import get_max_job_duration_limit
from paasta_tools.kubernetes.remote_run import get_remote_run_jobs
from paasta_tools.kubernetes.remote_run import get_remote_run_role_bindings
from paasta_tools.kubernetes.remote_run import get_remote_run_roles
from paasta_tools.kubernetes.remote_run import get_remote_run_service_accounts
from paasta_tools.kubernetes_tools import get_all_managed_namespaces
from paasta_tools.kubernetes_tools import KubeClient


ListingFuncType = Callable[[KubeClient, str], Sequence[Any]]
DeletionFuncType = Callable[[str, str], Any]
log = logging.getLogger(__name__)


def clean_namespace(
    kube_client: KubeClient,
    namespace: str,
    auth_age_limit: datetime,
    job_age_limit: datetime,
    dry_run: bool = False,
):
    """Clean ephemeral remote-run resource in a namespace

    :param KubeClient kube_client: kubernetes client
    :param str namepsace: kubernetes namespace
    :param datetime auth_age_limit: expiration time for authentication resources
    :param datetime job_age_limit: expiration time for job resources
    :param bool dry_run: delete resources for real or not
    """
    dry_run_msg = " (dry_run)" if dry_run else ""
    cleanup_actions: Sequence[Tuple[DeletionFuncType, ListingFuncType, datetime]] = (
        (
            kube_client.core.delete_namespaced_service_account,
            get_remote_run_service_accounts,
            auth_age_limit,
        ),
        (
            kube_client.rbac.delete_namespaced_role,
            get_remote_run_roles,
            auth_age_limit,
        ),
        (
            kube_client.rbac.delete_namespaced_role_binding,
            get_remote_run_role_bindings,
            auth_age_limit,
        ),
        (
            kube_client.batches.delete_namespaced_job,
            get_remote_run_jobs,
            job_age_limit,
        ),
    )
    for delete_func, list_func, age_limit in cleanup_actions:
        for entity in list_func(kube_client, namespace):
            if (
                not entity.metadata.name.startswith("remote-run-")
                or entity.metadata.creation_timestamp > age_limit
            ):
                continue
            log.info(f"Deleting {entity.metadata.name} in {namespace}{dry_run_msg}")
            if not dry_run:
                delete_func(entity.metadata.name, namespace)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Clean ephemeral Kubernetes resources created by remote-run invocations",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--max-age",
        type=int,
        default=600,
        help="Maximum age, in seconds, resources are allowed to have",
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        default=False,
        help="Do not actually delete resources",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="More verbose logging",
    )
    return parser.parse_args()


def main():
    args = parse_args()
    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.INFO)
    kube_client = KubeClient()
    now = datetime.now(tz=timezone.utc)
    age_limit = now - timedelta(seconds=args.max_age)
    job_age_limit = now - timedelta(seconds=get_max_job_duration_limit())
    for namespace in get_all_managed_namespaces(kube_client):
        clean_namespace(
            kube_client,
            namespace,
            auth_age_limit=age_limit,
            job_age_limit=job_age_limit,
            dry_run=args.dry_run,
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import copy
from typing import Any
from typing import cast
from typing import List
from typing import Optional

from mypy_extensions import TypedDict

from paasta_tools.long_running_service_tools import load_service_namespace_config
from paasta_tools.long_running_service_tools import LongRunningServiceConfig
from paasta_tools.long_running_service_tools import LongRunningServiceConfigDict
from paasta_tools.long_running_service_tools import ServiceNamespaceConfig
from paasta_tools.utils import BranchDictV2
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import Constraint  # noqa, imported for typing.
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import DockerParameter
from paasta_tools.utils import get_code_sha_from_dockerurl
from paasta_tools.utils import get_config_hash
from paasta_tools.utils import load_service_instance_config
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.utils import SystemPaastaConfig


MESOS_TASK_SPACER = "."


VolumeInfo = TypedDict(
    "VolumeInfo", {"container_path": str, "host_path": str, "mode": str}
)

_Docker_PortMapping = TypedDict(
    "_Docker_PortMapping", {"host_port": int, "container_port": int, "protocol": str}
)

DockerInfo = TypedDict(
    "DockerInfo",
    {
        "image": str,
        "network": str,
        "port_mappings": List[_Docker_PortMapping],
        "parameters": List[DockerParameter],
    },
)

ContainerInfo = TypedDict(
    "ContainerInfo", {"type": str, "docker": DockerInfo, "volumes": List[VolumeInfo]}
)


_CommandInfo_URI = TypedDict(
    "_CommandInfo_URI",
    {
        "value": str,
        "executable": bool,
        "extract": bool,
        "cache": bool,
        "output_file": str,
    },
    total=False,
)

CommandInfo = TypedDict(
    "CommandInfo",
    {
        "uris": List[_CommandInfo_URI],
        "environment": Any,
        "shell": bool,
        "value": str,
        "arguments": List[str],
        "user": str,
    },
    total=False,
)

Value_Scalar = TypedDict("Value_Scalar", {"value": float})

Value_Range = TypedDict("Value_Range", {"begin": int, "end": int})

Value_Ranges = TypedDict("Value_Ranges", {"range": List[Value_Range]})

Resource = TypedDict(
    "Resource",
    {"name": str, "type": str, "scalar": Value_Scalar, "ranges": Value_Ranges},
    total=False,
)


TaskID = TypedDict("TaskID", {"value": str})

SlaveID = TypedDict("SlaveID", {"value": str})


class TaskInfoBase(TypedDict):
    name: str
    task_id: TaskID
    agent_id: SlaveID
    resources: List[Resource]


class TaskInfo(TaskInfoBase):
    container: ContainerInfo
    command: CommandInfo


class NativeServiceConfigDict(LongRunningServiceConfigDict):
    pass


class NativeServiceConfig(LongRunningServiceConfig):
    config_dict: NativeServiceConfigDict
    config_filename_prefix = "paasta_native"

    def __init__(
        self,
        service: str,
        instance: str,
        cluster: str,
        config_dict: NativeServiceConfigDict,
        branch_dict: Optional[BranchDictV2],
        soa_dir: str,
        service_namespace_config: Optional[ServiceNamespaceConfig] = None,
    ) -> None:
        super().__init__(
            cluster=cluster,
            instance=instance,
            service=service,
            config_dict=config_dict,
            branch_dict=branch_dict,
            soa_dir=soa_dir,
        )
        # service_namespace_config may be omitted/set to None at first, then set
        # after initializing. e.g. we do this in load_paasta_native_job_config
        # so we can call get_nerve_namespace() to figure out what SNC to read.
        # It may also be set to None if this service is not in nerve.
        if service_namespace_config is not None:
            self.service_namespace_config = service_namespace_config
        else:
            self.service_namespace_config = ServiceNamespaceConfig()

    def task_name(self, base_task: TaskInfo) -> str:
        code_sha = get_code_sha_from_dockerurl(
            base_task["container"]["docker"]["image"]
        )

        filled_in_task = copy.deepcopy(base_task)
        filled_in_task["name"] = ""
        filled_in_task["task_id"] = {"value": ""}
        filled_in_task["agent_id"] = {"value": ""}

        config_hash = get_config_hash(
            filled_in_task, force_bounce=self.get_force_bounce()
        )

        return compose_job_id(
            self.service,
            self.instance,
            git_hash=code_sha,
            config_hash=config_hash,
            spacer=MESOS_TASK_SPACER,
        )

    def base_task(
        self, system_paasta_config: SystemPaastaConfig, portMappings=True
    ) -> TaskInfo:
        """Return a TaskInfo Dict with all the fields corresponding to the
        configuration filled in.

        Does not include task.agent_id or a task.id; those need to be
        computed separately.
        """
        docker_volumes = self.get_volumes(
            system_volumes=system_paasta_config.get_volumes(),
        )
        task: TaskInfo = {
            "name": "",
            "task_id": {"value": ""},
            "agent_id": {"value": ""},
            "container": {
                "type": "DOCKER",
                "docker": {
                    "image": self.get_docker_url(),
                    "parameters": [
                        {"key": param["key"], "value": param["value"]}
                        for param in self.format_docker_parameters()
                    ],
                    "network": self.get_mesos_network_mode(),
                    "port_mappings": [],
                },
                "volumes": [
                    {
                        "container_path": volume["containerPath"],
                        "host_path": volume["hostPath"],
                        "mode": volume["mode"].upper(),
                    }
                    for volume in docker_volumes
                ],
            },
            "command": {
                "value": str(self.get_cmd()),
                "uris": [
                    {
                        "value": system_paasta_config.get_dockercfg_location(),
                        "extract": False,
                    }
                ],
            },
            "resources": [
                {
                    "name": "cpus",
                    "type": "SCALAR",
                    "scalar": {"value": self.get_cpus()},
                },
                {"name": "mem", "type": "SCALAR", "scalar": {"value": self.get_mem()}},
            ],
        }

        if portMappings:
            task["container"]["docker"]["port_mappings"] = [
                {
                    "container_port": self.get_container_port(),
                    # filled by tasks_and_state_for_offer()
                    "host_port": 0,
                    "protocol": "tcp",
                }
            ]

            task["resources"].append(
                {
                    "name": "ports",
                    "type": "RANGES",
                    "ranges": {
                        # filled by tasks_and_state_for_offer
                        "range": [{"begin": 0, "end": 0}]
                    },
                }
            )

        task["name"] = self.task_name(task)

        return task

    def get_mesos_network_mode(self) -> str:
        return self.get_net().upper()


def load_paasta_native_job_config(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
    instance_type: str = "paasta_native",
    config_overrides: Optional[NativeServiceConfigDict] = None,
) -> NativeServiceConfig:
    instance_config_dict = cast(
        NativeServiceConfigDict,
        load_service_instance_config(
            service=service,
            instance=instance,
            instance_type=instance_type,
            cluster=cluster,
            soa_dir=soa_dir,
        ),
    )
    branch_dict: Optional[BranchDictV2] = None
    instance_config_dict.update(config_overrides or {})
    if load_deployments:
        deployments_json = load_v2_deployments_json(service, soa_dir=soa_dir)
        temp_instance_config = NativeServiceConfig(
            service=service,
            cluster=cluster,
            instance=instance,
            config_dict=instance_config_dict,
            branch_dict=None,
            soa_dir=soa_dir,
        )
        branch = temp_instance_config.get_branch()
        deploy_group = temp_instance_config.get_deploy_group()
        branch_dict = deployments_json.get_branch_dict(service, branch, deploy_group)

    service_config = NativeServiceConfig(
        service=service,
        cluster=cluster,
        instance=instance,
        config_dict=instance_config_dict,
        branch_dict=branch_dict,
        soa_dir=soa_dir,
    )

    service_namespace_config = load_service_namespace_config(
        service=service, namespace=service_config.get_nerve_namespace(), soa_dir=soa_dir
    )
    service_config.service_namespace_config = service_namespace_config

    return service_config


class UnknownNativeServiceError(Exception):
    pass



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./setup_kubernetes_cr.py [options]

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- -v, --verbose: Verbose output
"""
import argparse
import logging
import sys
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Sequence

from kubernetes.client.exceptions import ApiException

from paasta_tools import yaml_tools as yaml
from paasta_tools.cli.utils import LONG_RUNNING_INSTANCE_TYPE_HANDLERS
from paasta_tools.flink_tools import get_flink_ingress_url_root
from paasta_tools.kubernetes_tools import create_custom_resource
from paasta_tools.kubernetes_tools import CustomResourceDefinition
from paasta_tools.kubernetes_tools import ensure_namespace
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import KubeCustomResource
from paasta_tools.kubernetes_tools import KubeKind
from paasta_tools.kubernetes_tools import list_custom_resources
from paasta_tools.kubernetes_tools import load_custom_resource_definitions
from paasta_tools.kubernetes_tools import paasta_prefixed
from paasta_tools.kubernetes_tools import sanitise_kubernetes_name
from paasta_tools.kubernetes_tools import update_custom_resource
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_config_hash
from paasta_tools.utils import get_git_sha_from_dockerurl
from paasta_tools.utils import load_all_configs
from paasta_tools.utils import load_system_paasta_config


log = logging.getLogger(__name__)


class StdoutKubeClient:
    """Replace all destructive operations in Kubernetes APIs with
    writing out YAML to stdout."""

    class StdoutWrapper:
        def __init__(self, target) -> None:
            self.target = target

        def __getattr__(self, attr):
            if attr.startswith("create") or attr.startswith("replace"):
                return self.yaml_dump
            return getattr(self.target, attr)

        def yaml_dump(self, **kwargs):
            body = kwargs.get("body")
            if not body:
                return
            ns = kwargs.get("namespace")
            if ns:
                if "metadata" not in body:
                    body["metadata"] = {}
                body["metadata"]["namespace"] = ns
            yaml.safe_dump(body, sys.stdout, indent=4, explicit_start=True)

    def __init__(self, kube_client) -> None:
        self.deployments = StdoutKubeClient.StdoutWrapper(kube_client.deployments)
        self.core = StdoutKubeClient.StdoutWrapper(kube_client.core)
        self.policy = StdoutKubeClient.StdoutWrapper(kube_client.policy)
        self.apiextensions = StdoutKubeClient.StdoutWrapper(kube_client.apiextensions)
        self.custom = StdoutKubeClient.StdoutWrapper(kube_client.custom)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Creates custom_resources.")
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", dest="verbose", default=False
    )
    parser.add_argument(
        "-s", "--service", default=None, help="Service to setup CRs for"
    )
    parser.add_argument(
        "-i", "--instance", default=None, help="Service instance to setup CR for"
    )
    parser.add_argument(
        "-D",
        "--dry-run",
        action="store_true",
        default=False,
        help="Output kubernetes configuration instead of applying it",
    )
    parser.add_argument(
        "-c", "--cluster", default=None, help="Cluster to setup CRs for"
    )
    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()
    soa_dir = args.soa_dir
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    kube_client: Any = KubeClient()
    if args.dry_run:
        kube_client = StdoutKubeClient(kube_client)

    system_paasta_config = load_system_paasta_config()
    cluster = args.cluster or system_paasta_config.get_cluster()
    custom_resource_definitions = load_custom_resource_definitions(system_paasta_config)
    setup_kube_succeeded = setup_all_custom_resources(
        kube_client=kube_client,
        soa_dir=soa_dir,
        cluster=cluster,
        custom_resource_definitions=custom_resource_definitions,
        service=args.service,
        instance=args.instance,
    )
    sys.exit(0 if setup_kube_succeeded else 1)


def setup_all_custom_resources(
    kube_client: KubeClient,
    soa_dir: str,
    cluster: str,
    custom_resource_definitions: Sequence[CustomResourceDefinition],
    service: str = None,
    instance: str = None,
) -> bool:

    got_results = False
    succeeded = False
    # We support two versions due to our upgrade to 1.22
    # this functions runs succefully when any of the two apiextensions
    # succeed to update the CRDs as the cluster could be in any version
    # we need to try both possibilities
    for apiextension in [
        kube_client.apiextensions,
    ]:

        try:
            crds_list = apiextension.list_custom_resource_definition(
                label_selector=paasta_prefixed("service")
            ).items
        except ApiException:
            log.debug(
                "Listing CRDs with apiextensions/v1 not supported on this cluster"
            )
            crds_list = []

        cluster_crds = {crd.spec.names.kind for crd in crds_list}
        log.debug(f"CRDs found: {cluster_crds}")
        results = []
        for crd in custom_resource_definitions:
            if crd.kube_kind.singular not in cluster_crds:
                # TODO: kube_kind.singular seems to correspond to `crd.names.kind`
                # and not `crd.names.singular`
                log.warning(f"CRD {crd.kube_kind.singular} " f"not found in {cluster}")
                continue

            # by convention, entries where key begins with _ are used as templates
            # and will be filter out here
            config_dicts = load_all_configs(
                cluster=cluster, file_prefix=crd.file_prefix, soa_dir=soa_dir
            )

            ensure_namespace(
                kube_client=kube_client, namespace=f"paasta-{crd.kube_kind.plural}"
            )
            results.append(
                setup_custom_resources(
                    kube_client=kube_client,
                    kind=crd.kube_kind,
                    crd=crd,
                    config_dicts=config_dicts,
                    version=crd.version,
                    group=crd.group,
                    cluster=cluster,
                    service=service,
                    instance=instance,
                )
            )
        if results:
            got_results = True
            if any(results):
                succeeded = True
    # we want to return True if we never called `setup_custom_resources`
    # (i.e., we noop'd) or if any call to `setup_custom_resources`
    # succeed (handled above) - otherwise, we want to return False
    return succeeded or not got_results


def setup_custom_resources(
    kube_client: KubeClient,
    kind: KubeKind,
    version: str,
    crd: CustomResourceDefinition,
    config_dicts: Mapping[str, Mapping[str, Any]],
    group: str,
    cluster: str,
    service: str = None,
    instance: str = None,
) -> bool:
    succeded = True
    if config_dicts:
        crs = list_custom_resources(
            kube_client=kube_client, kind=kind, version=version, group=group
        )
    for svc, config in config_dicts.items():
        if service is not None and service != svc:
            continue
        if not reconcile_kubernetes_resource(
            kube_client=kube_client,
            service=svc,
            instance=instance,
            instance_configs=config,
            kind=kind,
            custom_resources=crs,
            version=version,
            group=group,
            cluster=cluster,
            crd=crd,
        ):
            succeded = False
    return succeded


def get_dashboard_base_url(kind: str, cluster: str, is_eks: bool) -> Optional[str]:
    system_paasta_config = load_system_paasta_config()
    dashboard_links = system_paasta_config.get_dashboard_links()
    if kind.lower() == "flink":
        flink_link = dashboard_links.get(cluster, {}).get("Flink")
        if flink_link is None:
            flink_link = get_flink_ingress_url_root(cluster, is_eks)
        if flink_link[-1:] != "/":
            flink_link += "/"
        return flink_link
    return None


def get_cr_owner(kind: str) -> Optional[str]:
    system_paasta_config = load_system_paasta_config()
    owners = system_paasta_config.get_cr_owners()
    return owners.get(kind.lower())


def format_custom_resource(
    instance_config: Mapping[str, Any],
    service: str,
    instance: str,
    cluster: str,
    kind: str,
    version: str,
    group: str,
    namespace: str,
    git_sha: str,
    is_eks: bool,
) -> Mapping[str, Any]:
    sanitised_service = sanitise_kubernetes_name(service)
    sanitised_instance = sanitise_kubernetes_name(instance)
    resource: Mapping[str, Any] = {
        "apiVersion": f"{group}/{version}",
        "kind": kind,
        "metadata": {
            "name": f"{sanitised_service}-{sanitised_instance}",
            "namespace": namespace,
            "labels": {
                "yelp.com/paasta_service": service,
                "yelp.com/paasta_instance": instance,
                "yelp.com/paasta_cluster": cluster,
                paasta_prefixed("service"): service,
                paasta_prefixed("instance"): instance,
                paasta_prefixed("cluster"): cluster,
            },
            "annotations": {},
        },
        "spec": instance_config,
    }
    if is_eks:
        resource["metadata"]["labels"][paasta_prefixed("eks")] = str(is_eks)

    url = get_dashboard_base_url(kind, cluster, is_eks)
    if url:
        resource["metadata"]["annotations"][paasta_prefixed("dashboard_base_url")] = url
    owner = get_cr_owner(kind)
    if owner:
        resource["metadata"]["labels"]["yelp.com/owner"] = owner
    config_hash = get_config_hash(resource)

    resource["metadata"]["annotations"]["yelp.com/desired_state"] = "running"
    resource["metadata"]["annotations"][paasta_prefixed("desired_state")] = "running"
    resource["metadata"]["labels"]["yelp.com/paasta_config_sha"] = config_hash
    resource["metadata"]["labels"][paasta_prefixed("config_sha")] = config_hash
    resource["metadata"]["labels"][paasta_prefixed("git_sha")] = git_sha
    return resource


def reconcile_kubernetes_resource(
    kube_client: KubeClient,
    service: str,
    instance_configs: Mapping[str, Any],
    custom_resources: Sequence[KubeCustomResource],
    kind: KubeKind,
    version: str,
    group: str,
    crd: CustomResourceDefinition,
    cluster: str,
    instance: str = None,
) -> bool:
    succeeded = True
    config_handler = LONG_RUNNING_INSTANCE_TYPE_HANDLERS[crd.file_prefix]

    is_eks = False
    if crd.file_prefix.endswith("eks"):
        is_eks = True

    for inst, config in instance_configs.items():
        if instance is not None and instance != inst:
            continue
        try:
            soa_config = config_handler.loader(
                service=service,
                instance=inst,
                cluster=cluster,
                load_deployments=True,
                soa_dir=DEFAULT_SOA_DIR,
            )
            git_sha = get_git_sha_from_dockerurl(soa_config.get_docker_url(), long=True)
            formatted_resource = format_custom_resource(
                instance_config=config,
                service=service,
                instance=inst,
                cluster=cluster,
                kind=kind.singular,
                version=version,
                group=group,
                namespace=f"paasta-{kind.plural}",
                git_sha=git_sha,
                is_eks=is_eks,
            )
            desired_resource = KubeCustomResource(
                service=service,
                instance=inst,
                config_sha=formatted_resource["metadata"]["labels"][
                    paasta_prefixed("config_sha")
                ],
                git_sha=formatted_resource["metadata"]["labels"].get(
                    paasta_prefixed("git_sha")
                ),
                kind=kind.singular,
                name=formatted_resource["metadata"]["name"],
                namespace=f"paasta-{kind.plural}",
            )
            if not (service, inst, kind.singular) in [
                (c.service, c.instance, c.kind) for c in custom_resources
            ]:
                log.info(f"{desired_resource} does not exist so creating")
                create_custom_resource(
                    kube_client=kube_client,
                    version=version,
                    kind=kind,
                    formatted_resource=formatted_resource,
                    group=group,
                )
            elif desired_resource not in custom_resources:
                sanitised_service = sanitise_kubernetes_name(service)
                sanitised_instance = sanitise_kubernetes_name(inst)
                log.info(f"{desired_resource} exists but config_sha doesn't match")
                update_custom_resource(
                    kube_client=kube_client,
                    name=f"{sanitised_service}-{sanitised_instance}",
                    version=version,
                    kind=kind,
                    formatted_resource=formatted_resource,
                    group=group,
                )
            else:
                log.info(f"{desired_resource} is up to date, no action taken")
        except Exception as e:
            log.error(str(e))
            succeeded = False
    return succeeded


if __name__ == "__main__":
    main()
